/*:ts=8*/
/*****************************************************************************
 * FIDOGATE --- Gateway UNIX Mail/News <-> FIDO NetMail/EchoMail
 *
 * $Id: config.c,v 3.7 1994/05/02 05:36:42 mj Exp mj $
 *
 * Configuration data and functions
 *
 *****************************************************************************
 * Copyright (C) 1990, 1993
 *  _____ _____
 * |     |___  |   Martin Junius             FIDO:      2:242/6.1
 * | | | |   | |   Republikplatz 3           Internet:  mj@sungate.fido.de
 * |_|_|_|@home|   D-52072 Aachen, Germany   Phone:     ++49-241-86931 (voice)
 *
 * This file is part of FIDOGATE.
 *
 * FIDOGATE is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * FIDOGATE is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with FIDOGATE; see the file COPYING.  If not, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************/

#include "fidogate.h"



/*
 * Current line in config file
 */
static int scf_line;



/*
 * Directories
 */
static char scf_libdir[MAXPATH];
static char scf_spooldir[MAXPATH];
static char scf_outbound[MAXPATH];
static char scf_inbound[MAXPATH];

/*
 * Hostname / domainname
 */
static char scf_hostname[MAXPATH];
static char scf_domainname[MAXPATH];
static char scf_hostsdomain[MAXPATH];
static char scf_fqdn[MAXPATH];

/*
 * EchoMail origin line
 */
static char scf_origin[MAXPATH];

/*
 * News organization
 */
static char scf_organization[MAXPATH];



/*
 * Configured zones and addresses
 */
struct st_addr
{
    int zone;				/* Zone for this address set */
    Node addr;				/* Our own main address */
    Node fake;				/* Our own fakenet address */
    Node uplink;			/* Uplink address */
};

static struct st_addr scf_addr[MAXADDRESS];
static int scf_naddr = 0;		/* # of addresses */
static int scf_zone  = 0;		/* Current zone */
static int scf_index = 0;		/* Index to current zone scf_addr[] */


/*
 * Zones, domains, and outbound directories
 */
struct st_zones
{
    int zone;				/* Zone */
    char *inet_domain;			/* Internet domain */
    char *ftn_domain;			/* FTN domain */
    char *out;				/* Outbound subdir */
};

static struct st_zones scf_zones[MAXADDRESS];
static int scf_nzones = 0;		/* # of zones */
static int scf_izones = 0;		/* Index for cf_zones_trav() */


/*
 * Configured DOS drive -> UNIX path translation
 */
struct st_dos
{
    char *drive;			/* MSDOS drive */
    char *path;				/* UNIX path */
};

static struct st_dos scf_dos[MAXDOSDRIVE];
static int scf_ndos = 0;		/* # of DOS drives */


/*
 * FTN-Internet gateway
 */
static Node scf_gateway;



/*
 * Debug output of configuration
 */
void cf_debug()
{
    int i;

    debug(4, "libdir=%s  spooldir=%s", scf_libdir, scf_spooldir);
    debug(4, "inbound=%s outbound=%s", scf_inbound, scf_outbound);
    debug(4, "fqdn=%s", scf_fqdn);
    debug(4, "origin=%s", scf_origin);
    debug(4, "organization=%s", scf_organization);
    
    for(i=0; i<scf_naddr; i++)
	debug(4, "Address Z%-4d: %-16s %-16s Uplink: %s",
	      scf_addr[i].zone,
	      node_to_asc(&scf_addr[i].addr, TRUE),
	      node_to_asc(&scf_addr[i].fake, TRUE),
	      node_to_asc(&scf_addr[i].uplink, TRUE)          );

    for(i=0; i<scf_nzones; i++)
	debug(4, "Zone %-4d: %s  %s  %s", scf_zones[i].zone,
	      scf_zones[i].inet_domain, scf_zones[i].ftn_domain,
	      scf_zones[i].out                                 );
#ifdef ROUTE_VIA_FTN_GATEWAY
    debug(4, "gateway=%s", node_to_asc(&scf_gateway, TRUE));
#endif
}



/*
 * Return current main/fakenet/uplink FTN address
 */
Node *cf_addr()
{
    return &scf_addr[scf_index].addr;
}

Node *cf_fake()
{
    return &scf_addr[scf_index].fake;
}

Node *cf_uplink()
{
    return &scf_addr[scf_index].uplink;
}

Node cf_n_addr()
{
    return scf_addr[scf_index].addr;
}

Node cf_n_fake()
{
    return scf_addr[scf_index].fake;
}

Node cf_n_uplink()
{
    return scf_addr[scf_index].uplink;
}



/*
 * Set current zone
 */
void cf_set_zone(zone)
int zone;
{
    int i;
    
    if(scf_naddr == 0) 
    {
	fprintf(stderr, "No FTN addresses configured.\n");
	exit(1);
    }

    scf_zone = zone;
    for(i=0; i<scf_naddr; i++)
	if(zone == scf_addr[i].zone)
	{
	    scf_index = i;
	    debug(3, "Select Z%d AKA: %-16s %-16s Uplink: %s",
		  scf_addr[i].zone,
		  node_to_asc(&scf_addr[i].addr, TRUE),
		  node_to_asc(&scf_addr[i].fake, TRUE),
		  node_to_asc(&scf_addr[i].uplink, TRUE)          );
	    return;
	}
    
    scf_index = i = 0;
    debug(3, "Select default AKA: %-16s %-16s Uplink: %s",
	  node_to_asc(&scf_addr[i].addr, TRUE),
	  node_to_asc(&scf_addr[i].fake, TRUE),
	  node_to_asc(&scf_addr[i].uplink, TRUE)          );
}



/*
 * Return current/default zone
 */
int cf_zone()
{
    return scf_zone;
}

int cf_defzone()
{
    return scf_addr[0].zone;
}



/*
 * Read line from config file. Strip `\n', leading spaces,
 * comments (starting with `#'), and empty lines. cf_getline() returns
 * a pointer to the first non-whitespace in buffer.
 */
char *cf_getline(buffer, len, fp)
    char *buffer;
    int len;
    FILE *fp;
{
    char *p;

    scf_line = 0;
    while (fgets(buffer, len, fp)) {
	scf_line++;
	/* Strip LF or CR/LF at end of line */
	if(buffer[strlen(buffer)-1] == '\n')
	    buffer[strlen(buffer) - 1] = 0;
	if(buffer[strlen(buffer)-1] == '\r')
	    buffer[strlen(buffer) - 1] = 0;
	if((p = strchr(buffer, '#')))		/* Strip comments */
	    *p = 0;
	for(p=buffer; *p && isspace(*p); p++) ;	/* Skip white spaces */
	if (*p)
	    return p;
    }
    return NULL;
}



/*
 * Read config file
 */
void cf_read_config_file(name)
    char *name;
{
    FILE *cf;
    char *line, *p, *keyword;
    int ia=0, ir=0;

    if(name && !*name)			/* Empty string -> no config file */
	return;
    
    cf = name ? xfopen(name, R_MODE) : libfopen(CONFIG, R_MODE);
    
    while((line = cf_getline(buffer, BUFFERSIZE, cf)))
    {
	keyword = xstrtok(line, " \t");
	if(!keyword)
	    continue;

	/***** hostname *****************************************************/
	if	(!stricmp(keyword, "hostname"))
	{
	    p = xstrtok(NULL, " \t");
	    if(!p) 
	    {
		log("config: missing hostname");
		continue;
	    }
	    strncpy0(scf_hostname, p, MAXPATH);
	}
	/***** domain *******************************************************/
	else if (!stricmp(keyword, "domain"  ))
	{
	    p = xstrtok(NULL, " \t");
	    if(!p) 
	    {
		log("config: missing domainname");
		continue;
	    }
	    if(p[0] != '.')
	    {
		strncpy0(scf_domainname, ".", sizeof(scf_domainname));
		strncat0(scf_domainname, p  , sizeof(scf_domainname));
	    }
	    else
		strncpy0(scf_domainname, p  , sizeof(scf_domainname));
	    /* This is also the default for "HostsDomain" */
	    strncpy0(scf_hostsdomain, scf_domainname, sizeof(scf_hostsdomain));
	}
	/***** hostsdomain **************************************************/
	else if (!stricmp(keyword, "hostsdomain"  ))
	{
	    p = xstrtok(NULL, " \t");
	    if(!p) 
	    {
		log("config: missing domainname");
		continue;
	    }
	    if(p[0] != '.')
	    {
		strncpy0(scf_hostsdomain, ".", sizeof(scf_hostsdomain));
		strncat0(scf_hostsdomain, p  , sizeof(scf_hostsdomain));
	    }
	    else
		strncpy0(scf_hostsdomain, p  , sizeof(scf_hostsdomain));
	}
	/***** origin *******************************************************/
	else if (!stricmp(keyword, "origin"  ))
	{
	    p = xstrtok(NULL, " \t");
	    if(!p) 
	    {
		log("config: missing origin");
		continue;
	    }
	    strncpy0(scf_origin, p, MAXPATH);
	}
	/***** organization *************************************************/
	else if (!stricmp(keyword, "organization"))
	{
	    p = xstrtok(NULL, " \t");
	    if(!p) 
	    {
		log("config: missing organization");
		continue;
	    }
	    strncpy0(scf_organization, p, MAXPATH);
	}
	/***** libdir *******************************************************/
	else if (!stricmp(keyword, "libdir"))
	{
	    p = xstrtok(NULL, " \t");
	    if(!p) 
	    {
		log("config: missing libdir");
		continue;
	    }
	    strncpy0(scf_libdir, p, MAXPATH);
	}
	/***** spooldir *****************************************************/
	else if (!stricmp(keyword, "spooldir"))
	{
	    p = xstrtok(NULL, " \t");
	    if(!p) 
	    {
		log("config: missing spooldir");
		continue;
	    }
	    strncpy0(scf_spooldir, p, MAXPATH);
	}
	/***** outbound *****************************************************/
	else if (!stricmp(keyword, "outbound"))
	{
	    p = xstrtok(NULL, " \t");
	    if(!p) 
	    {
		log("config: missing outbound");
		continue;
	    }
	    strncpy0(scf_outbound, p, MAXPATH);
	}
	/***** inbound ******************************************************/
	else if (!stricmp(keyword, "inbound"))
	{
	    p = xstrtok(NULL, " \t");
	    if(!p) 
	    {
		log("config: missing inbound");
		continue;
	    }
	    strncpy0(scf_inbound, p, MAXPATH);
	}
	/***** address ******************************************************/
	else if (!stricmp(keyword, "address" ))
	{
	    Node a, f;
	    
	    /* Main address */
	    p = xstrtok(NULL, " \t");
	    if(!p) 
	    {
		log("config: missing address");
		continue;
	    }
	    if( asc_to_node(p, &a, FALSE) == ERROR )
	    {
		log("config: illegal address %s", p);
		continue;
	    }
	    /* Fakenet address */
	    p = xstrtok(NULL, " \t");
	    if(p)
	    {
		if( asc_to_node(p, &f, FALSE) == ERROR )
		{
		    log("config: illegal address %s", p);
		    continue;
		}
	    }
	    else
		f = a;

	    if(a.zone != f.zone)
	    {
		log("config: different zones in address");
		continue;
	    }

	    if(ia < MAXADDRESS)
	    {
		debug(5, "config: zone=%d addr=%s fake=%s",
		      a.zone, node_to_asc(&a, TRUE), node_to_asc(&f, TRUE));
		scf_addr[ia].zone   = a.zone;
		scf_addr[ia].addr   = a;
		scf_addr[ia].fake   = f;
		ia++;
	    }
	    else
		log("config: too many addresses");
	}
	/***** uplink *******************************************************/
	else if (!stricmp(keyword, "uplink" ))
	{
	    Node a;
	    
	    /* Main address */
	    p = xstrtok(NULL, " \t");
	    if(!p) 
	    {
		log("config: missing address");
		continue;
	    }
	    if( asc_to_node(p, &a, FALSE) == ERROR )
	    {
		log("config: illegal address %s", p);
		continue;
	    }

	    if(ir < MAXADDRESS)
	    {
		debug(5, "config: uplink=%s", node_to_asc(&a, TRUE));
		scf_addr[ir].uplink = a;
		ir++;
	    }
	    else
		log("config: too many addresses");
	}
	/***** zone *********************************************************/
	else if (!stricmp(keyword, "zone" ))
	{
	    int zone;
	    char *inet, *ftn, *out;

	    if(scf_nzones >= MAXADDRESS)
	    {
		log("config: too many zones");
		continue;
	    }
	    
	    if(! (p = xstrtok(NULL, " \t")) )
	    {
		log("config: missing zone");
		continue;
	    }
	    if(!stricmp(p, "default"))
		zone = 0;
	    else 
	    {
		zone = atoi(p);
		if(!zone)
		{
		    log("config: illegal zone value %s", p);
		    continue;
		}
	    }
	    
	    if(! (inet = xstrtok(NULL, " \t")) )
	    {
		log("config: missing Internet domain");
		continue;
	    }

	    if(! (ftn = xstrtok(NULL, " \t")) )
	    {
		log("config: missing FTN domain");
		continue;
	    }

	    if(! (out = xstrtok(NULL, " \t")) )
	    {
		log("config: missing outbound directory");
		continue;
	    }

	    scf_zones[scf_nzones].zone        = zone;
	    scf_zones[scf_nzones].inet_domain = strsave(inet);
	    scf_zones[scf_nzones].ftn_domain  = strsave(ftn);
	    scf_zones[scf_nzones].out         = strsave(out);
	    scf_nzones++;
	}
	/***** zone *********************************************************/
	else if (!stricmp(keyword, "dosdrive" ))
	{
	    char *drive, *path;
	    
	    if(scf_ndos >= MAXDOSDRIVE)
	    {
		log("config: too many DOS drives");
		continue;
	    }
	    if(! (drive = xstrtok(NULL, " \t")) )
	    {
		log("config: missing DOS drive");
		continue;
	    }
	    if(! (path = xstrtok(NULL, " \t")) )
	    {
		log("config: missing UNIX path");
		continue;
	    }

	    scf_dos[scf_ndos].drive = strsave(drive);
	    scf_dos[scf_ndos].path  = strsave(path);
	    scf_ndos++;
	}
	/***** gateway ******************************************************/
	else if (!stricmp(keyword, "gateway" ))
	{
	    Node a;
	    
	    /* Main address */
	    p = xstrtok(NULL, " \t");
	    if(!p) 
	    {
		log("config: missing address");
		continue;
	    }
	    if( asc_to_node(p, &a, FALSE) == ERROR )
	    {
		log("config: illegal address %s", p);
		continue;
	    }
	    scf_gateway = a;
	}
	/********************************************************************/
#if 0
	else 
	    log("config: unknown command %s", keyword);
#endif
    }

    if(ir != ia)
	log("config: # address != # uplink");
    scf_naddr = ia;
    scf_zone  = scf_addr[0].zone;
    scf_index = 0;

    strncpy0(scf_fqdn, scf_hostname,   MAXPATH);
    strncat0(scf_fqdn, scf_domainname, MAXPATH);
    
    fclose(cf);
}




/*
 * Init configuration
 */
void cf_initialize()
{
    strncpy0(scf_libdir,   LIBDIR,   MAXPATH);
    strncpy0(scf_spooldir, SPOOLDIR, MAXPATH);
}



/*
 * Set FIDO address
 */
void cf_set_addr(addr)
char *addr;
{
    Node node;
    
    if( asc_to_node(addr, &node, FALSE) == ERROR )
    {
	fprintf(stderr, "Illegal FIDO address %s\n", addr);
	exit(EX_USAGE);
    }

    scf_naddr = 1;
    scf_addr[0].zone = node.zone;
    scf_addr[0].addr = node;
    scf_addr[0].fake = node;
    scf_zone  = node.zone;
    scf_index = 0;
}



/*
 * Set FIDO fakenet address
 */
void cf_set_fake(addr)
char *addr;
{
    Node node;
    
    if( asc_to_node(addr, &node, FALSE) == ERROR )
    {
	fprintf(stderr, "Illegal FIDO address %s\n", addr);
	exit(EX_USAGE);
    }
    
    scf_naddr = 1;
    scf_addr[0].fake = node;
    scf_zone  = scf_addr[0].zone;
    scf_index = 0;
}



/*
 * Set uplink FIDO address
 */
void cf_set_uplink(addr)
char *addr;
{
    Node node;
    
    if( asc_to_node(addr, &node, FALSE) == ERROR )
    {
	fprintf(stderr, "Illegal FIDO address %s\n", addr);
	exit(EX_USAGE);
    }
    
    scf_naddr = 1;
    scf_addr[0].uplink = node;
    scf_zone  = scf_addr[0].zone;
    scf_index = 0;
}



/*
 * Set/get lib/spool directory
 */
void cf_set_libdir(dir)
char *dir;
{
    strncpy0(scf_libdir, dir, MAXPATH);
}

void cf_set_spooldir(dir)
char *dir;
{
    strncpy0(scf_spooldir, dir, MAXPATH);
}

char *cf_libdir()
{
    return scf_libdir;
}

char *cf_spooldir()
{
    return scf_spooldir;
}



/*
 * Return hostname / domain name / fully qualified domain name
 */
char *cf_hostname()
{
    return scf_hostname;
}

char *cf_domainname()
{
    return scf_domainname;
}

char *cf_hostsdomain()
{
    return scf_hostsdomain;
}

char *cf_fqdn()
{
    return scf_fqdn;
}


/*
 * Return origin
 */
char *cf_origin()
{
    return scf_origin;
}


/*
 * Return organization
 */
char *cf_organization()
{
    return scf_organization;
}


/*
 * Set/get outbound/inbound
 */
void cf_set_outbound(dir)
    char *dir;
{
    strncpy0(scf_outbound, dir, MAXPATH);
}

char *cf_outbound()
{
    return scf_outbound;
}

void cf_set_inbound(dir)
    char *dir;
{
    strncpy0(scf_inbound, dir, MAXPATH);
}

char *cf_inbound()
{
    return scf_inbound;
}



/***** Stuff for processing zone info ****************************************/

/*
 * Return Internet domain for a FIDO zone
 */
char *cf_zones_inet_domain(zone)
int zone;
{
    int i;
    
    /*
     * Search zone
     */
    for(i=0; i<scf_nzones; i++)
	if(scf_zones[i].zone == zone)
	    return scf_zones[i].inet_domain;

    /*
     * Search default domain
     */
    for(i=0; i<scf_nzones; i++)
	if(scf_zones[i].zone == 0)
	    return scf_zones[i].inet_domain;

    return FIDODOMAIN;
}



/*
 * Check for valid zone
 */
int cf_zones_check(zone)
int zone;
{
    int i;
    
    /*
     * Search zone
     */
    for(i=0; i<scf_nzones; i++)
	if(scf_zones[i].zone == zone)
	    return TRUE;

    return FALSE;    
}



/*
 * Traverse Internet domains
 */
char *cf_zones_trav(first)
int first;
{
    if(first)
	scf_izones = 0;

    return scf_izones < scf_nzones
	   ? scf_zones[scf_izones++].inet_domain
	   : NULL;
}



/*
 * Return outbound directory for a FIDO zone
 */
char *cf_zones_out(zone)
int zone;
{
    int i;
    
    /*
     * Search zone
     */
    for(i=0; i<scf_nzones; i++)
	if(scf_zones[i].zone == zone)
	    return scf_zones[i].out;

    return NULL;
}



/*
 * Traverse FTN addresses
 */
Node *cf_addr_trav(first)
    int first;
{
    static int iaddr, ifake;
    Node *node;
    
    if(first)
	iaddr = ifake = 0;

    node = NULL;
    do
    {
	if(iaddr >= scf_naddr)		/* End of addresses */
	    return NULL;

	if(!ifake) 			/* Return normal address */
	{
	    node = &scf_addr[iaddr].addr;
	    ifake = 1;
	}
	else
	{
	    /** If fakenet address is the same as the normal one, skip it **/
	    if(!node_eq(&scf_addr[iaddr].addr, &scf_addr[iaddr].fake))
		node = &scf_addr[iaddr].fake;
	    iaddr++;
	    ifake = 0;
	}
    }
    while(node == NULL);
    
    return node;
}



/*
 * UNIX to MSDOS file name translation enabled
 */
int cf_dos()
{
    return scf_ndos != 0;
}



/*
 * Convert UNIX path on server to MSDOS path on client
 */
char *cf_dos_xlate(name)
    char *name;
{
    static char buf[MAXPATH];
    int i;
    char *s;
    int len;
    
    for(i=0; i<scf_ndos; i++)
    {
	len = strlen(scf_dos[i].path);
	if(!strncmp(name, scf_dos[i].path, len))
	{
	    strncpy0(buf, scf_dos[i].drive, MAXPATH);
	    strncat0(buf, name+len        , MAXPATH);
	    for(s=buf; *s; s++)
		switch(*s)
		{
		case '/':	*s = '\\';  break;
		}
	    return buf;
	}
    }
    
    return NULL;
}



/*
 * Address of FTN-Internet gateway
 */
Node cf_gateway()
{
    return scf_gateway;
}
