/*:ts=8*/
/*****************************************************************************
 * FIDOGATE --- Gateway UNIX Mail/News <-> FIDO NetMail/EchoMail
 *
 * $Id: fido2rfc.c,v 3.7 1994/05/02 05:36:42 mj Exp mj $
 *
 * Unpack FIDO mail packets in INBOUND directory and convert to RFC mails
 * and news batches (directories INBOUND_MAIL / INBOUND_NEWS)
 *
 *****************************************************************************
 * Copyright (C) 1990, 1993
 *  _____ _____
 * |     |___  |   Martin Junius             FIDO:      2:242/6.1
 * | | | |   | |   Republikplatz 3           Internet:  mj@sungate.fido.de
 * |_|_|_|@home|   D-52072 Aachen, Germany   Phone:     ++49-241-86931 (voice)
 *
 * This file is part of FIDOGATE.
 *
 * FIDOGATE is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * FIDOGATE is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with FIDOGATE; see the file COPYING.  If not, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************/

#include "fidogate.h"
#include "getopt.h"

#include <pwd.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/wait.h>



#define PROGRAM "fido2rfc"
#define VERSION "$Revision: 3.7 $"



/*
 * Prototypes
 */
int 	check_origin		P((char *, Node *));
void	check_ctrl_a		P((char *));
int	check_rfc_header	P((char *));
char   *msg_getline		P((char *, int, FILE *));
Area   *news_msg		P((void));
int	msg_format_buffer	P((char *, Textlist *));
int	unpack			P((FILE *, Packet *));
int	unpack_file		P((char *));

void	short_usage		P((void));
void	usage			P((void));



/*
 * For recognizing `^AINTL', `^AFMPT x', `^AMSGID ...' and `^AREPLY ...'
 */
#define SAVEBUFSIZ 256

static int  msgbody_fmpt = 0;
static char msgbody_area[SAVEBUFSIZ];
static char msgbody_msgid[SAVEBUFSIZ];
static char msgbody_reply[SAVEBUFSIZ];
static char msgbody_intl[SAVEBUFSIZ];
static char msgbody_origid[SAVEBUFSIZ];
static char msgbody_origref[SAVEBUFSIZ];


/*
 * For grabbing addresses from RFC822 header lines at start of message
 */

static char *msgbody_rfc_from;
static char *msgbody_rfc_to;


/*
 * Command line options
 */
int t_flag = FALSE;

char in_dir[BUFSIZ];
char news_dir[BUFSIZ];
char news_name[BUFSIZ];
char news_tmp[BUFSIZ];
FILE *news_file;



/*
 * check_origin() --- Analyse ` * Origin: ...' line in FIDO message
 *		      body and parse address in ()s into Node structure
 *
 * Origin line is checked for the rightmost occurence of
 * ([text] z:n/n.p).
 */

int check_origin(buffer, node)
char *buffer;
Node *node;
{
    char *left, *right;
    char *buf;

    if(!strncmp(" * Origin:", buffer, 10)) {
	buf   = strsave(buffer);
	right = strrchr(buf, ')');
	if(!right) {
	    free(buf);
	    return(FALSE);
	}
	left  = strrchr(buf, '(');
	if(!left) {
	    free(buf);
	    return(FALSE);
	}
	*right = 0;
	*left++ = 0;

	/*
	 * Parse node info
	 */
	while(*left && !isdigit(*left))
	    left++;
	if(asc_to_node(left, node, FALSE) != OK) {
	    /* Not a valid FIDO address */
	    node->zone = node->net = node->node = node->point = -1;
	    free(buf);
	    return(FALSE);
	}
	free(buf);
	return(TRUE);
    }
    return(FALSE);
}



/*
 * check_ctrl_a() --- Check for various ^A kludges
 */
void check_ctrl_a(buffer)
char *buffer;
{
    char *p;

    if(buffer[strlen(buffer) - 1] == '\n')
	buffer[strlen(buffer) - 1] = 0;		/* Strip trailing \n */
    
    /*
     * Look for `^AINTL ...'
     */
    if(!strncmp(buffer+1, "INTL", 4)) {
	buffer += 5;
	while(*buffer==':' || *buffer==' ')
	    buffer++;
	strcpy(msgbody_intl, buffer);
	return;
    }

    /*
     * Look for `^AFMPT x'
     */
    if(!strncmp(buffer+1, "FMPT", 4)) {
	for(p=buffer+5; *p && !isdigit(*p); p++);
	if(*p)
	    msgbody_fmpt = atoi(p);
	return;
    }
    
    /*
     * Look for `^AMSGID: ...'
     */
    if(!strncmp(buffer+1, "MSGID: ", 7)) {
	strcpy(msgbody_msgid, buffer+8);
	return;
    }
    
    /*
     * Look for `^AREPLY: ...'
     */
    if(!strncmp(buffer+1, "REPLY: ", 7)) {
	strcpy(msgbody_reply, buffer+8);
	return;
    }

    /*
     * Look for `^AORIGID: ...'
     */
    if(!strncmp(buffer+1, "ORIGID: ", 8)) {
	strcpy(msgbody_origid, buffer+9);
	return;
    }

    /*
     * Look for `^AORIGREF: ...'
     */
    if(!strncmp(buffer+1, "ORIGREF: ", 9)) {
	strcpy(msgbody_origref, buffer+10);
	return;
    }

    /*
     * Look for `^ACHARSET ...' or `^ACHRS ...'
     */
    if(!strncmp(buffer+1, "CHARSET:", 8)) {
	charset_set(buffer + 9);
	return;
    }
    if(!strncmp(buffer+1, "CHRS:", 5)) {
	charset_set(buffer + 6);
	return;
    }
}



/*
 * check_rfc_header() --- Check for RFC822 like header lines at
 *			  the beginning of the FIDO message body
 */

int check_rfc_header(buffer)
char *buffer;
{
    if(!strnicmp(buffer, "From: ", 6)) {
	msgbody_rfc_from = strsaveline(buffer + 6);
	return(1);
    }
    if(!strnicmp(buffer, "Reply-To: ", 10)) {
	msgbody_rfc_from = strsaveline(buffer + 10);
	return(1);
    }
    if(!strncmp(buffer, "UUCPFROM:", 9)) {
	msgbody_rfc_from = strsaveline(buffer + 9);
	return(1);
    }
    if(!strnicmp(buffer, "To: ", 4)) {
	msgbody_rfc_to = strsaveline(buffer + 4);
	return(1);
    }
    if(!strnicmp(buffer, "To:", 3)) {
	msgbody_rfc_to = strsaveline(buffer + 3);
	return(1);
    }
    return(0);
}



/*
 * Read lines from text body of FIDO message.
 */

char *msg_getline(buffer, size, fp)
char *buffer;
int size;
FILE *fp;
{
    int c=0;
    char *cp;
    char *xl;

    /*
     * Read lines from message body. Lines are terminated by \r, \n and
     * 0x8d (soft \r) are ignored. High bit characters are converted according
     * to the active character set, control character are converted to ^X.
     * (Except ^A at the beginning of a line)
     */
    cp = buffer;

    while (--size>1 && (c = getc(fp)) != EOF)
    {
	if(c==1 && cp==buffer)		/* ^A at beginning of line */
	{
	    *cp++ = c;
	    continue;
	}
	
#if 1
	/*
	 * Special handling for a bug in SQUISH 1.01
	 * KLUDGE!!!
	 */
	if(c == 2)
	{
	    c = getc(fp);
	    if(c == 0)
	    {
		long pos = ftell(fp);
		
		fseek(fp, pos-2, 0);
		log("Bad packet: SQUISH bug!");
		fprintf(stderr, "Bad packet: SQUISH bug!\n");
		break;
	    }
	}
#endif	    
	
	/*
	 * Check for orphan 0-Byte in message, next Byte must be 0 or 2
	 */
	if(c == 0) 
	{
	    c = getc(fp);
	    ungetc(c, fp);
	    if(c==2 || c==0)
	    {
		c = 0;
		break;
	    }
	    c = 0;
	}
	
	if(c=='\n' || c==0x8d)		/* Ignore \n and soft \r */
	    continue;
	if(c == '\r')
	    c = '\n';
	else if(c < ' ') {
	    /*
	     * Substitute control chars with '^X'
	     */
	    if(c != '\t')
	    {
		size--;
		*cp++ = '^';
		c = c + '@';
	    }
	}
	else if(c & 0x80)
	{
	    /*
	     * Translate special characters according to character set
	     */
	    xl = charset_xlate(c);
	    if(!xl || !*xl)
		continue;
	    while(xl[1])
	    {
		size--;
		*cp++ = *xl++;
	    }
	    c = *xl;
	}
	
	*cp++ = c;
	if (c == '\n')			/* End of line */
	    break;
    }

    /*
     * 0 byte == end of message
     * Put it back, if it occurred in the middle of a line
     */
    if(c==0 && cp!=buffer)
    {
	*cp++ = '\n';			/* Add newline */
	ungetc(0, fp);
    }

    *cp = 0;

    return (c==0 || c==EOF) && cp==buffer ? NULL : buffer;
}



/*
 * Test for news article and returns newsgroup for FIDO area in msgbody_area[]
 */
Area *news_msg()
{
    char *p;
    Area *pa;
    static Area area;
    
    if(msgbody_area[0])
    {
	/* Message is FIDO EchoMail */
	msgbody_area[strlen(msgbody_area) - 1] = 0;
	for(p=msgbody_area+strlen("AREA:"); *p && isspace(*p); p++);
	debug(3, "FIDO Area: %s", p);
	pa = areas_lookup(p, NULL);
	if(pa)
	{
	    debug(5, "Found: %s %s Z%d", pa->area, pa->group, pa->zone);
	    return pa;
	}
	area.next  = NULL;
	area.area  = p;
	area.group = NULL;
	area.zone  = 0;
	return &area;
    }

    /* Message is FIDO NetMail */
    return NULL;
}



/*
 * Format buffer line and put it into Textlist. Returns number of
 * lines.
 */
int msg_format_buffer(buffer, tlist)
char *buffer;
Textlist *tlist;
{
    /*
     * Remark: MAX_LINELEN+16 and MAX_LINELEN+8, the numbers +16 and +8
     *         are just arbitrary values to be safe.
     */
    char *p, *np;
    char localbuffer[MAX_LINELEN + 16];
    int i;
    int lines;
    
    if(strlen(buffer) <= MAX_LINELEN)		/* Nothing to do */
    {
	Textlist_append(tlist, buffer);
	return 1;
    }
    else
    {
	/*
	 * Break line with word wrap
	 */
	lines = 0;
	p     = buffer;

	while(TRUE)
	{
	    /*
	     * Search backward for a whitespace to break line. If no proper
	     * point is found, the line will not be splitted.
	     */
	    for(i=MAX_LINELEN-1; i>=0; i--)
		if(isspace(p[i]))	/* Found a white space */
		    break;
	    if(i == -1)				/* Not found, put line as is */
	    {
		Textlist_append(tlist, p);
		lines++;
		return lines;
	    }
	    for(; i>=0 && isspace(p[i]); i--);	/* Skip more white space */
	    i++;				/* Return to last white sp. */

	    /*
	     * Cut here and put into textlist
	     */
	    np = p + i;
	    *np++ = 0;
	    strncpy0(localbuffer, p, MAX_LINELEN+8);
	    strcat(localbuffer, "\n");
	    Textlist_append(tlist, localbuffer);
	    lines++;
	    
	    /*
	     * Advance buffer pointer and test length of remaining line
	     */
	    p = np;
	    for(; *p && isspace(*p); p++);	/* Skip white space */
	    if(*p == 0)				/* The end */
		return lines;
	    if(strlen(p) <= MAX_LINELEN)	/* No more wrappin' */
	    {
		Textlist_append(tlist, p);
		lines++;
		return lines;
	    }

	    /* Play it again, Sam! */
	}
    }
    /**NOT REACHED**/
    return 0;
}



/*
 * Unpack FIDO mail packet. Each message header und some information in
 * the message body (^A kludges and some RFC like lines) ist converted
 * to an RFC compatible message header. If the first line of the message
 * starts with `AREA:' then message is processed as a news article, else
 * as a mail message.
 */

int unpack(pkt_file, pkt)
    FILE *pkt_file;
    Packet *pkt;
{
    Message msg;			/* Message header */
    RFCAddr addr_from, addr_to;		/* Representation of FTN addresses */
    
    char mail_to[MAXINETADDR];		/* Addressee of mail */
    char x_orig_to[MAXINETADDR];

    Node origin_node;			/* Address in origin line */
    int lines;				/* Lines in message body */
    char *p;
    int messagetype;
    int c;
    Node node;
    int rfc_header_flag, area_header_flag, origin_flag;
    char *thisdomain, *uplinkdomain;
    char *from_line, *to_line=NULL;	/* From:/To: for output */
    char *id_line, *ref_line;		/* Message-ID:/References: */
    int this_is_mail, this_is_news;	/* Flags */
    int uucp_flag;
    char *tear_line=NULL, *save_tear_line=NULL;
    char *origin_line=NULL;
    Area *pa;
    Textlist theader;
    Textlist tbody;    
    Textlist tvia;
    Textlist tpath;


    Textlist_init(&theader);
    Textlist_init(&tbody);
    Textlist_init(&tvia);
    Textlist_init(&tpath);

    news_file = NULL;


    while ((messagetype = pkt_get_int16(pkt_file)) == MSG_TYPE) {
	/*
	 * Clear stuff set by parsing message body
	 */
	origin_node.zone   = -1;
	msgbody_area[0]    = 0;
	msgbody_fmpt	   = 0;
	msgbody_msgid[0]   = 0;
	msgbody_reply[0]   = 0;
	msgbody_intl[0]	   = 0;
	msgbody_origid[0]  = 0;
	msgbody_origref[0] = 0;
	if(msgbody_rfc_from)
	    free(msgbody_rfc_from);
	msgbody_rfc_from = NULL;
	if(msgbody_rfc_to)
	    free(msgbody_rfc_to);
	msgbody_rfc_to = NULL;
	if(tear_line)
	    free(tear_line);
	tear_line = NULL;
	if(origin_line)
	    free(origin_line);
	origin_line = NULL;
	
	lines = 0;
	Textlist_clear(&theader);
	Textlist_clear(&tbody);
	Textlist_clear(&tvia);
	Textlist_clear(&tpath);

	charset_reset();
	x_orig_to[0] = 0;

	msg.node_from = pkt->from;
	msg.node_to   = pkt->to;
	pkt_get_msg_hdr(pkt_file, &msg);
	
	if (!*msg.subject)
	    strcpy(msg.subject, "(no subject)");

	/*
	 * Read entire FIDO message body and store it in Textlist tbody
	 * for later output. (Must read message body first because we need
	 * some information from it.)
	 * Some special treatment for origin line and ^A
	 * kludges is done.
	 */
	rfc_header_flag  = TRUE;
	area_header_flag = TRUE;
	origin_flag      = FALSE;
	tear_line        = NULL;
	save_tear_line   = NULL;
	while(msg_getline(buffer, BUFFERSIZE, pkt_file))
	{
	    if(area_header_flag && !strncmp("AREA:", buffer, 5))
	    {					/* AREA:TAG EchoMail */
		area_header_flag = FALSE;
		strncpy0(msgbody_area, buffer, sizeof(msgbody_area));
		continue;
	    }		
		
	    if(buffer[0] == 1) {		/* ^A kludge line */
		if(!strncmp("\001Via ", buffer, 5))
		    Textlist_appendf(&tvia, "X-FTN-Via: %s", buffer+5);
		else
		    if(!strncmp("\001PATH: ", buffer, 7))
		        Textlist_appendf(&tpath, "X-FTN-Path: %s", buffer+7);
		    else
		        check_ctrl_a(buffer);
		continue;
	    }

	    area_header_flag = FALSE;

	    if(rfc_header_flag) {		/* RFC header line */
		if((rfc_header_flag = check_rfc_header(buffer)))
		    continue;
	    }

	    if(!strncmp(" * Origin: ", buffer, 11)) /* Origin line */
	    {
		if(save_tear_line)
		{
		    tear_line      = save_tear_line;
		    save_tear_line = NULL;
		}
		check_origin(buffer, &origin_node);
		origin_line = strsave(buffer + 11);
		origin_flag = TRUE;
		continue;
	    }

	    /*
	     * If there is a save tear line now, it really isn't one,
	     * put it into body text list.
	     */
	    if(save_tear_line)
	    {
		lines += msg_format_buffer(save_tear_line, &tbody);
		free(save_tear_line);
		save_tear_line = NULL;
	    }
	    
	    if(!strncmp("---", buffer, 3))	/* Possible tear line */
	    {
		save_tear_line = strsave(buffer);
		continue;
	    }
		
	    if( origin_flag && !strncmp("SEEN-BY:", buffer, 8) )
		continue;

	    lines += msg_format_buffer(buffer, &tbody);
	}

	/*
	 * The last "---" line in a message is also a tear line
	 */
	if(save_tear_line)
	    tear_line = save_tear_line;
	

	/*
	 * Check for mail or news.
	 */
	if((pa = news_msg()))
	{
	    if(!news_file)
	    {
		long n = sequencer(SEQ_NEWS);

		sprintf(news_tmp,  "%s/%08ld.tmp", news_dir, n);
		sprintf(news_name, "%s/%08ld.msg", news_dir, n);
		news_file = fopen(news_tmp, W_MODE);
		if(!news_file)
		{
		    log("$Can't create output file %s", news_tmp);
		    return ERROR;
		}
	    }

	    this_is_news = TRUE;
	    this_is_mail = FALSE;
	}
	else
	{
	    /*
	     * This is personal mail
	     */

	    this_is_mail = TRUE;
	    this_is_news = FALSE;
	}

	/*
	 * Construct real from address, using information from
	 * header, origin line and ^A kludges.
	 */
	if(this_is_news && origin_node.zone != -1) {
	    msg.node_orig = origin_node;
	}
	else {
	    if(*msgbody_intl) {
		p = strtok(msgbody_intl, " \n");    /* Destination */
		if(p && asc_to_node(p, &node, FALSE) == OK)
		    msg.node_to = node;
		
		p = strtok(NULL	   , " \n");	    /* Source */
		if(p && asc_to_node(p, &node, FALSE) == OK)
		    msg.node_from = node;

		msg.node_orig = msg.node_from;
		    
	    }
	    if(msgbody_fmpt) {
		msg.node_orig.point = msgbody_fmpt;
	    }
	}
	debug(2, "FIDO From: %s", node_to_asc(&msg.node_orig, TRUE));

	/*
	 * Set current node and AKA address according to sender's zone
	 */
	cf_set_zone(msg.node_orig.zone);

	/*
	 * Convert FTN from/to addresses to RFCAddr struct
	 */
	addr_from = rfcaddr_from_ftn(msg.name_from, &msg.node_orig);
	addr_to   = rfcaddr_from_ftn(msg.name_to,   &msg.node_to  );

	uucp_flag = FALSE;
	if(!stricmp(addr_to.real, "UUCP")    ||
	   !stricmp(addr_to.real, "GATEWAY")   )
	{
	    /* Don't output UUCP or GATEWAY as real name */
	    uucp_flag = TRUE;
	    addr_to.real[0] = 0;
	}
	

	/*
	 * Do mail alias checking on both from and to names
	 */
	if(!msgbody_rfc_from)
	{
	    Alias *a;

	    debug(4, "Checking for alias: %s",
		  rfcaddr_to_asc(&addr_from, TRUE));
	    a = alias_lookup(&msg.node_orig, NULL, addr_from.real);
	    if(a)
	    {
		debug(3, "Alias found: %s %s %s", a->username,
		      node_to_asc(&a->node, FALSE), a->fullname);
		strncpy0(addr_from.user, a->username, sizeof(addr_from.user));
	    }
	}
	if(this_is_mail && !msgbody_rfc_to)
	{
	    Alias *a;
	    
	    debug(4, "Checking for alias: %s",
		  rfcaddr_to_asc(&addr_to, TRUE));
	    a = alias_lookup(cf_addr(), NULL, addr_to.real);
	    if(a)
	    {
		debug(3, "Alias found: %s %s %s", a->username,
		      node_to_asc(&a->node, FALSE), a->fullname);
		strcpy(mail_to, a->username);
	    }
	    else
		strcpy(mail_to, addr_to.user);
	}

	/*
	 * Special handling for -u flag (unsecure):
	 *   Messages with To line will be bounced
	 */
	if(this_is_mail && t_flag && msgbody_rfc_to)
	{
	    debug(1, "Unsecure message with To line");
	    log("Bouncing mail from %s", rfcaddr_to_asc(&addr_from, TRUE));
	    bounce_mail("unsecure", &addr_from, &msg, msgbody_rfc_to, &tbody);
	    goto message_done;
	}
	    
	/*
	 * There are message trackers out there in FIDONET. Obviously
	 * they can't handling addressing the gateway so we send mail
	 * from "MsgTrack..." etc. to postmaster.
	 */
	if(    !strnicmp(addr_from.user, "MsgTrack", 8)
	    || !strnicmp(addr_from.user, "Reflex_Netmail_Policeman", 24)
	    || !strnicmp(addr_from.user, "TrackM", 6)
	    || !strnicmp(addr_from.user, "ITrack", 6)
	    || !strnicmp(addr_from.user, "O/T-Track", 9)
	 /* || whatever ... */		    )
	{
	    debug(1, "Mail from FIDO message tracker");
	    strcpy(x_orig_to, mail_to);
	    strcpy(mail_to, POSTMASTER);
	    /* Reset uucp_flag to avoid bouncing of these messages */
	    uucp_flag = FALSE;
	}


	thisdomain   = ftn_to_inet(cf_addr(),       TRUE);
	uplinkdomain = ftn_to_inet(&msg.node_from,  TRUE);

#ifdef HOSTS_RESTRICTED
	/*
	 * Bounce mail from nodes not within domain .fido.de
	 */
	if(this_is_mail && msgbody_rfc_to)
	    if(hosts_lookup(&msg.node_orig, NULL) == NULL)
	    {
		/*
		 * Not registered in HOSTS
		 */
		debug(1, "Not a Fido.DE node: %s",
		      node_to_asc(&msg.node_orig, FALSE));
		log("Bouncing mail from %s", rfcaddr_to_asc(&addr_from, TRUE));
		bounce_mail("restricted", &addr_from, &msg,
			    msgbody_rfc_to, &tbody);
		goto message_done;
	    }
#endif /**HOSTS_RESTRICTED**/

	/*
	 * Check mail messages' user name
	 */
	if(this_is_mail && !msgbody_rfc_to &&
	   !strchr(mail_to, '@') && !strchr(mail_to, '%') &&
	   !strchr(mail_to, '!')			    )
	{
	    if(uucp_flag) 
	    {
		/*
		 * Addressed to `UUCP' or `GATEWAY', but no To: line
		 */
		debug(1, "Message to `UUCP' or `GATEWAY' without To line");
		log("Bouncing mail from %s", rfcaddr_to_asc(&addr_from, TRUE));
		bounce_mail("noto", &addr_from, &msg, msgbody_rfc_to, &tbody);
		goto message_done;
	    }
#ifdef ADD_DOMAIN
	    /*
	     * Add `@host.domain' to local address
	     */
	    strcat(mail_to, "@");
	    strcat(mail_to, cf_fqdn());
#endif
	}

	/*
	 * Construct string for From: header line
	 */
	if(msgbody_rfc_from) {
	    if(strchr(msgbody_rfc_from, '(') || strchr(msgbody_rfc_from, '<'))
		from_line = buf_sprintf("%s", msgbody_rfc_from);
	    else
		from_line = buf_sprintf("%s (%s)", msgbody_rfc_from,
					addr_from.real);
	}
	else
	{
	    from_line = rfcaddr_to_asc(&addr_from, TRUE);
	}

	/*
	 * Construct string for To:/X-Comment-To: header line
	 */
	if(msgbody_rfc_to) {
	    if(strchr(msgbody_rfc_to, '(') ||
	       strchr(msgbody_rfc_to, '<') || !*addr_to.real)
		to_line = buf_sprintf("%s", msgbody_rfc_to);
	    else
		to_line = buf_sprintf("%s (%s)", msgbody_rfc_to, addr_to.real);
	}
	else {
	    if(this_is_news)
	    {
		if(!strcmp(addr_to.user, "All")  ||
		   !strcmp(addr_to.user, "Alle") ||
		   !strcmp(addr_to.user, "*.*")  ||
		   !strcmp(addr_to.user, "*")      )
		    to_line = NULL;
		else
		    to_line = buf_sprintf("(%s)", addr_to.real);
	    }
	    
	    if(this_is_mail)
		to_line = buf_sprintf("%s", mail_to);
	}

	/*
	 * Construct Message-ID and References header lines
	 */
	id_line  = NULL;
	ref_line = NULL;
	
	if(*msgbody_origid)
	    id_line = msgid_convert_origid(msgbody_origid, TRUE);
	if(!id_line)
	{
	    id_line = msgid_fido_to_rfc(msgbody_msgid);
	    if(id_line == NULL)
		id_line = msgid_default(&msg.node_orig, msg.name_from,
					msg.name_to,
					msg.subject, msg.date);
	}
	if(*msgbody_origref)
	    ref_line = msgid_convert_origid(msgbody_origref, FALSE);
	if(!ref_line)
	    ref_line = msgid_fido_to_rfc(msgbody_reply);

	/*
	 * Output RFC mail/news header
	 *
	 * Mail
	 */
	if(this_is_mail) {
	    /*
	     * Log it
	     */
	    log("%-13s  %s -> %s", node_to_asc(&msg.node_orig, TRUE),
		from_line, to_line);
	    
	    if(mail_open() == ERROR)
		return ERROR;

	    Textlist_appendf(&theader,
			     "From %s %s\n", rfcaddr_to_asc(&addr_from, FALSE),
			     date("%a %b %d %H:%M:%S %Y", NULL) );
	    Textlist_appendf(&theader,
		"Received: by %s (FIDOGATE %s)\n",
		thisdomain, version_global()                            );
	    Textlist_appendf(&theader,
		"\tid AA%05d; %s\n",
		getpid(), date(NULL, NULL) );
	    Textlist_appendf(&theader,
		"Date: %s\n",
		date(DATE_MAIL, &msg.date)			);
	    Textlist_appendf(&theader,
		"From: %s\n",
		from_line						);
	    Textlist_appendf(&theader,
		"Subject: %s\n",
		msg.subject						);
	    Textlist_appendf(&theader,
		"Message-ID: %s\n",
		id_line							);
	    Textlist_appendf(&theader,
		"To: %s\n",
		to_line							);
	    if(*x_orig_to)
	    Textlist_appendf(&theader,
		"X-Orig-To: %s\n",
		x_orig_to						);
#ifdef ERRORS_TO
	    Textlist_appendf(&theader,
		"Errors-To: %s\n",
		ERRORS_TO						);
#endif
	    Textlist_appendf(&theader,
			     "Organization: %s\n", cf_organization() );
	    Textlist_appendf(&theader,
		"Lines: %d\n",
		lines							);
	    Textlist_appendf(&theader,
		"X-Gateway: FIDOGATE %s\n", version_global()		);
#ifdef X_FTN_ORIGIN
	    if(origin_line)
		Textlist_appendf(&theader, "X-FTN-Origin: %s", origin_line );
#endif
#ifdef X_FTN_TEARLINE
	    if(tear_line)
		Textlist_appendf(&theader, "X-FTN-Tearline:%s", tear_line+3 );
#endif
#ifdef X_FTN_VIA
	    Textlist_add(&theader, &tvia);
#endif
	}

	/*
	 * News
	 */
	if(this_is_news) {
	    Textlist_appendf(&theader,
		"Path: %s!%s!not-for-mail\n",
		thisdomain, uplinkdomain		);
	    Textlist_appendf(&theader,
		"Date: %s\n",
		date(DATE_NEWS, &msg.date)			);
	    Textlist_appendf(&theader,
		"From: %s\n",
		from_line						);
	    Textlist_appendf(&theader,
		"Subject: %s\n",
		msg.subject						);
	    Textlist_appendf(&theader,
		"Message-ID: %s\n",
		id_line							);
	    if(ref_line)
	    Textlist_appendf(&theader,
		"References: %s\n",
		ref_line						);
	    Textlist_appendf(&theader,
		"Newsgroups: %s\n",
		pa->group ? pa->group : FTN_JUNK			);
	    if(pa->distribution)
		Textlist_appendf(&theader, "Distribution: %s\n",
				 pa->distribution               );
	    if(to_line)
		Textlist_appendf(&theader, "X-Comment-To: %s\n", to_line);
	    Textlist_appendf(&theader,
			     "Organization: %s\n", cf_organization() );
	    Textlist_appendf(&theader, "Lines: %d\n", lines);
	    Textlist_appendf(&theader,
		"X-Gateway: FIDOGATE %s\n", version_global()		);
#ifdef X_FTN_ORIGIN
	    if(origin_line)
		Textlist_appendf(&theader, "X-FTN-Origin: %s", origin_line );
#endif
#ifdef X_FTN_TEARLINE
	    if(tear_line)
		Textlist_appendf(&theader, "X-FTN-Tearline:%s", tear_line+3 );
#endif
#ifdef X_FTN_PATH
	    Textlist_add(&theader, &tpath);
#endif
	    if(!pa->group)
		Textlist_appendf(&theader, "X-FTN-Area: %s\n", pa->area );
	}

	Textlist_appendf(&theader, "\n");

	/*
	 * Write header and message body to output file
	 */
	if(this_is_news) {
	    /* News batch */
	    fprintf(news_file, "#! rnews %ld\n", Textlist_size(&theader) +
						Textlist_size(&tbody)	);
	    Textlist_print(&theader, news_file);
	    Textlist_print(&tbody,   news_file);
	}
	if(this_is_mail) {
	    /* Mail message */
	    Textlist_print(&theader, mail_file());
	    Textlist_print(&tbody,   mail_file());
	    /* Close mail */
	    mail_close();
	}

message_done:
	/*
	 * Dome with this message.
	 */
	debug(1, "Done with message");
	continue;

	/*
	 * In case of error skip text of message
	 */
	while((c = getc(pkt_file)) && c != EOF);
    }

    if(news_file) 
    {
	/* Close news file */
	fclose(news_file);
	/* Rename .tmp -> .msg */
	if(rename(news_tmp, news_name) == -1)
	    log("$Can't rename %s to %s", news_tmp, news_name);
    }
    
    if (messagetype != MSG_TYPE && messagetype != EOF && messagetype != 0) {
	log("Strange ending: %d", messagetype);
	return -1;
    }

    debug(1, "Done with packet");

    return 0;
}



/*
 * Unpack one packet file
 */
int unpack_file(pkt_name)
    char *pkt_name;
{
    Packet pkt;
    FILE *pkt_file;

    debug(1, "Unpacking FTN packet %s", pkt_name);

    /*
     * Open packet and read header
     */
    pkt_file = fopen(pkt_name, R_MODE);
    if(!pkt_file) {
	log("$Can't open packet %s", pkt_name);
	exit(EX_OSERR);
    }
    if(pkt_get_hdr(pkt_file, &pkt) == ERROR)
    {
	log("Error reading header from %s", pkt_name);
	exit(EX_OSERR);
    }
    
    /*
     * Unpack it
     */
    if(unpack(pkt_file, &pkt) == -1) 
    {
	log("Error in unpacking %s", pkt_name);
	exit(EX_OSERR);
    }
    
    fclose(pkt_file);
    
    if (unlink(pkt_name)) {
	log("$Could not unlink packet %s", pkt_name);
	exit(EX_OSERR);
    }

    return OK;
}



/*
 * Usage messages
 */
void short_usage()
{
    fprintf(stderr, "usage: %s [-options] [packet ...]\n", PROGRAM);
    fprintf(stderr, "       %s --help  for more information\n", PROGRAM);
}


void usage()
{
    fprintf(stderr, "FIDOGATE %s  %s %s\n\n",
	    version_global(), PROGRAM, version_local(VERSION) );
    
    fprintf(stderr, "usage:   %s [-options] [packet ...]\n\n", PROGRAM);
    fprintf(stderr, "\
options: -u --unsecure                toss unsecure packets\n\
         -I --inbound-dir name        set inbound packets directory\n\
         -l --lock-file name          create lock file while processing\n\
         -x --exec-program name       exec program after processing\n\
\n\
	 -v --verbose                 more verbose\n\
	 -h --help                    this help\n\
         -c --config name             read config file (\"\" = none)\n\
	 -L --lib-dir name            set lib directory\n\
	 -S --spool-dir name          set spool directory\n\
	 -a --addr Z:N/F.P            set FTN address\n\
	 -f --fake-addr Z:N/F.P       set FTN fakenet address\n\
	 -u --uplink-addr Z:N/F.P     set FTN uplink address\n\
         -z --zone Z                  select aka for zone Z\n"  );
}



/***** main() ****************************************************************/

int main(argc, argv)
    int argc;
    char *argv[];
{
    struct dirent *dir;
    DIR *dp;
    int c;
    char *lockname = NULL;
    char *execprog = NULL;
    char *I_flag=NULL;
    char *c_flag=NULL;
    char *S_flag=NULL, *L_flag=NULL;
    char *a_flag=NULL, *f_flag=NULL, *u_flag=NULL;
    char *z_flag=NULL;
    char pkt_name[MAXPATH];

    int option_index;
    static struct option long_options[] =
    {
	{ "unsecure",     0, 0, 't'},	/* Toss unsecure packets */
	{ "inbound-dir",  1, 0, 'I'},	/* Set inbound packets directory */
	{ "lock-file",    1, 0, 'l'},	/* Create lock file while processing */
	{ "exec-program", 1, 0, 'x'},	/* Exec program after processing */

	{ "verbose",      0, 0, 'v'},	/* More verbose */
	{ "help",         0, 0, 'h'},	/* Help */
	{ "config",       1, 0, 'c'},	/* Config file */
	{ "spool-dir",    1, 0, 'S'},	/* Set FIDOGATE spool directory */
	{ "lib-dir",      1, 0, 'L'},	/* Set FIDOGATE lib directory */
	{ "addr",         1, 0, 'a'},	/* Set FIDO address */
	{ "fake-addr",    1, 0, 'f'},	/* Set FIDO fakenet address */
	{ "uplink-addr",  1, 0, 'u'},	/* Set FIDO uplink address */
	{ "zone",         1, 0, 'z'},	/* Select aka for zone */
	{ 0,              0, 0, 0  }
    };

    /* Init configuration */
    cf_initialize();


    while ((c = getopt_long(argc, argv, "tI:l:x:vhc:S:L:a:f:u:z:",
			    long_options, &option_index     )) != EOF)
	switch (c) {
	/***** fido2rfc options *****/
	case 't':
	    /* Unsecure */
	    t_flag = TRUE;
	    break;
	case 'I':
	    /* Inbound packets directory */
	    I_flag = optarg;
	    break;
        case 'l':
            /* Lock file */
            lockname = optarg;
            break;
        case 'x':
            /* Exec program after unpack */
            execprog = optarg;
            break;

	/***** Common options *****/
	case 'v':
	    verbose++;
	    break;
	case 'h':
	    usage();
	    exit(0);
	    break;
	case 'c':
	    c_flag = optarg;
	    break;
	case 'S':
	    S_flag = optarg;
	    break;
	case 'L':
	    L_flag = optarg;
	    break;
	case 'a':
	    a_flag = optarg;
	    break;
	case 'f':
	    f_flag = optarg;
	    break;
	case 'u':
	    u_flag = optarg;
	    break;
	case 'z':
	    z_flag = optarg;
	    break;
	default:
	    short_usage();
	    exit(EX_USAGE);
	    break;
	}

    /*
     * Read config file
     */
    if(L_flag)				/* Must set libdir beforehand */
	cf_set_libdir(L_flag);
    cf_read_config_file(c_flag);

    /*
     * Process config options
     */
    if(L_flag)
	cf_set_libdir(L_flag);
    if(S_flag)
	cf_set_spooldir(S_flag);
    if(a_flag)
	cf_set_addr(a_flag);
    if(f_flag)
	cf_set_fake(f_flag);
    if(u_flag)
	cf_set_uplink(u_flag);

    cf_debug();

    if(z_flag)
	cf_set_zone(atoi(z_flag));
    
    /*
     * Process local options
     */
    if(I_flag)
    {
	strncpy0(in_dir, I_flag, sizeof(in_dir));
	sprintf(mail_dir, "%s/%s", in_dir, INBOUND_MAIL);
	sprintf(news_dir, "%s/%s", in_dir, INBOUND_NEWS);
    }
    else 
    {
	/* Default inbound directory */
	if(t_flag)
	    sprintf(in_dir, "%s/%s", cf_spooldir(), UNSECURE);
	else
	    sprintf(in_dir, "%s/%s", cf_spooldir(), INBOUND);
	sprintf(mail_dir, "%s/%s/%s", cf_spooldir(), INBOUND, INBOUND_MAIL);
	sprintf(news_dir, "%s/%s/%s", cf_spooldir(), INBOUND, INBOUND_NEWS);
    }
    
    /*
     * Init various modules
     */
    areas_init();
    hosts_init();
    alias_init();
#ifdef MAUS_DE
    maus_init();
#endif

    /*
     * If called with -l lock option, try to create lock FILE
     */
    if(lockname)
    {
	int fd;

	/*
	 * Use open() with flag O_EXCL, this will fail if the
	 * lock file already exists
	 */
	fd = open(buf_sprintf("%s/%s", cf_spooldir(), lockname),
		  O_RDWR | O_CREAT | O_EXCL, 0666);
	debug(1, "lock %s", fd==-1 ? "failed" : "succeeded");
	if(fd == -1)			    /* Failed */
	    exit(2);
	close(fd);
    }

    if(optind >= argc)
    {
	/*
	 * No packet file args, process in directory
	 */
	dp = opendir(in_dir);
	
	if(!dp)
	{
	    log("$Unable to open inbound directory %s", in_dir);
	    if(lockname)
		unlink(lockname);
	    exit(EX_OSERR);
	}
    
	while((dir = readdir(dp)))
	{
	    if(wildmat(dir->d_name, "*.pkt"))
	    {
		sprintf(pkt_name, "%s/%s", in_dir, dir->d_name);
		unpack_file(pkt_name);
	    }
	}
	
	closedir(dp);
    }
    else
    {
	/*
	 * Process packet files on command line
	 */
	for(; optind<argc; optind++)
	    unpack_file(argv[optind]);
    }
    

    /*
     * Execute given command, if option -x set.
     */
    if(execprog)
    {
	char buf[128];
	
	sprintf(buf, "%s/%s", cf_libdir(), execprog);
	system(buf);
    }
    
    if(lockname)
	if(unlink(buf_sprintf("%s/%s", cf_spooldir(), lockname)))
	    log("$Error removing lockfile");
    
    exit(EX_OK);
}
