/*:ts=8*/
/*****************************************************************************
 * FIDOGATE --- Gateway UNIX Mail/News <-> FIDO NetMail/EchoMail
 *
 * $Id: packet.c,v 3.7 1994/05/02 05:36:42 mj Exp mj $
 *
 * Functions to read/write packets and messages
 *
 *****************************************************************************
 * Copyright (C) 1990, 1993
 *  _____ _____
 * |     |___  |   Martin Junius             FIDO:      2:242/6.1
 * | | | |   | |   Republikplatz 3           Internet:  mj@sungate.fido.de
 * |_|_|_|@home|   D-52072 Aachen, Germany   Phone:     ++49-241-86931 (voice)
 *
 * This file is part of FIDOGATE.
 *
 * FIDOGATE is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * FIDOGATE is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with FIDOGATE; see the file COPYING.  If not, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************/

#include "fidogate.h"



/*
 * Read null-terminated string from packet file
 */
int pkt_get_string(fp, buf, nbytes)
    FILE *fp;
    char *buf;
    int nbytes;
{
    int c, i;

    for(i=0; TRUE; i++)
    {
	c = getc(fp);
	if(c==0 || c==EOF)
	    break;
	if(i >= nbytes-1)
	    break;
	buf[i] = c;
    }
    buf[i] = 0;
    
    return c!=0 ? ERROR : OK;
}



/*
 * Return date of message in UNIX format (secs after the epoche)
 *
 * Understood formats: see FTS-0001 (actually using getdate.y parser)
 */
time_t pkt_get_date(fp)
    FILE *fp;
{
    char buf[MSG_MAXDATE];

    /*
     * Treat FTN date as a 0-terminated string. Actually it's a fixed
     * string with exactly 19 chars + 0-char.
     */
    buf[0] = 0;
    pkt_get_string(fp, buf, sizeof(buf));
    
    return parsedate(buf, NULL);
}



/*
 * Read message header from packet file
 */
int pkt_get_msg_hdr(fp, msg)
    FILE *fp;
    Message *msg;
{
    msg->node_from.node = pkt_get_int16(fp);
    msg->node_to  .node = pkt_get_int16(fp);
    msg->node_from.net  = pkt_get_int16(fp);
    msg->node_to  .net  = pkt_get_int16(fp);
    msg->node_orig      = msg->node_from;
    msg->attr           = pkt_get_int16(fp);
    msg->cost           = pkt_get_int16(fp);
    msg->date           = pkt_get_date(fp);

    pkt_get_string( fp, msg->name_to  , sizeof(msg->name_to  ) );
    pkt_get_string( fp, msg->name_from, sizeof(msg->name_from) );
    pkt_get_string( fp, msg->subject  , sizeof(msg->subject  ) );

    debug(2, "FTN message header:");
    debug(2, "    From: %-36s @ %s",
	  msg->name_from, node_to_asc(&msg->node_from, TRUE));
    debug(2, "    To  : %-36s @ %s",
	  msg->name_to  , node_to_asc(&msg->node_to  , TRUE));
    debug(2, "    Subj: %s", msg->subject);
    debug(2, "    Date: %s",
	  msg->date!=-1 ? date(NULL, &msg->date) : "LOCAL" );
    debug(3, "    Attr: %04x", msg->attr);
    
    return ferror(fp);
}



/*
 * Write string to packet in null-terminated format.
 */
int pkt_put_string(fp, s)
FILE *fp;
char *s;
{
    fputs(s, fp);
    putc(0, fp);
    
    return ferror(fp);
}



/*
 * Write line to packet, replacing \n with \r\n
 */
int pkt_put_line(fp, s)
FILE *fp;
char *s;
{
    for(; *s; s++)
    {
	if(*s == '\n')
	    putc('\r', fp);
	putc(*s, fp);
    }

    return ferror(fp);
}
    


/*
 * Write 16-bit integer in 80x86 format, i.e. low byte first,
 * then high byte. Machine independent function.
 */
int pkt_put_int16(fp, value)
FILE *fp;
int value;
{
    putc(value & 0xff, fp);
    putc((value >> 8) & 0xff, fp);

    return ferror(fp);
}



/*
 * Write date/time in FTS-0001 format
 */
int pkt_put_date(pkt, t)
    FILE *pkt;
    time_t t;
{
    static time_t last = -1;

    if(t == -1)
    {
	/* No valid time, use local time */
	debug(5, "Invalid time - using local time");
	t = time(NULL);
	/* Kludge to avoid the same date/time */
	if(t == last)
	    t += 2;
	last = t;
    }
    
    debug(9, "t = %ld", (long)t);

    /* Date according to FTS-0001 */
    
    pkt_put_string(pkt, date("%d %b %y  %H:%M:%S", &t) );

    return ferror(pkt);
}



/*
 * Write message header to packet.
 */
int pkt_put_msg_hdr(pkt, msg, kludge_flag)
    FILE *pkt;
    Message *msg;
    int kludge_flag;		/* TRUE: write AREA/^AINTL,^AFMPT,^ATOPT */
{
    debug(2, "FTN message header:");
    debug(2, "    From: %-36s @ %s",
	  msg->name_from, node_to_asc(&msg->node_from, TRUE));
    debug(2, "    To  : %-36s @ %s",
	  msg->name_to  , node_to_asc(&msg->node_to  , TRUE));
    debug(2, "    Subj: %s", msg->subject);
    debug(2, "    Date: %s",
	  msg->date!=-1 ? date(NULL, &msg->date) : "LOCAL" );
    debug(3, "    Attr: %04x", msg->attr);

    /*
     * Write message header
     */
    pkt_put_int16 (pkt, MSG_TYPE           );
    pkt_put_int16 (pkt, msg->node_from.node);
    pkt_put_int16 (pkt, msg->node_to  .node);
    pkt_put_int16 (pkt, msg->node_from.net );
    pkt_put_int16 (pkt, msg->node_to  .net );
    pkt_put_int16 (pkt, msg->attr          );
    pkt_put_int16 (pkt, msg->cost          );
    
    pkt_put_date  (pkt, msg->date          );
    pkt_put_string(pkt, msg->name_to       );
    pkt_put_string(pkt, msg->name_from     );
    pkt_put_string(pkt, msg->subject       );

    if(!kludge_flag)
	return ferror(pkt);
    
    /*
     * Write area tag / zone, point adressing kludges
     */
    if(msg->area)
	fprintf(pkt, "AREA:%s\r\n", msg->area);
    else 
    {
#ifdef FTN_FORCE_INTL
	if(TRUE)
#else
	if(cf_zone()    != msg->node_from.zone  ||
	   cf_defzone() != msg->node_from.zone  ||
	   cf_zone()    != msg->node_to  .zone  ||
	   cf_defzone() != msg->node_to  .zone     )
#endif
	{
	    Node tmpf, tmpt;
	    
	    tmpf = msg->node_from; tmpf.point = 0; tmpf.domain[0] = 0;
	    tmpt = msg->node_to;   tmpt.point = 0; tmpt.domain[0] = 0;
	    fprintf(pkt, "\001INTL %s %s\r\n",
		    node_to_asc(&tmpt, FALSE), node_to_asc(&tmpf, FALSE));
	}

/* Kludge: always generate ^AFMPT, because SQUISH adds point address,
 * if the origin address of the packet is a point address.
 **	if(msg->node_from.point)  **/
	    fprintf(pkt, "\001FMPT %d\r\n", msg->node_from.point);
	if(msg->node_to  .point)
	    fprintf(pkt, "\001TOPT %d\r\n", msg->node_to  .point);
    }

    return ferror(pkt);
}



/*
 * Read 16-bit integer in 80x86 format, i.e. low byte first,
 * then high byte. Machine independent function.
 */
long pkt_get_int16(fp)
FILE *fp;
{
    int c;
    unsigned val;

    if((c = getc(fp)) == EOF)
	return ERROR;
    val	 = c;
    if((c = getc(fp)) == EOF)
	return ERROR;
    val |= c << 8;

    return val;
}



/*
 * Read n bytes from file stream
 */
int pkt_get_nbytes(fp, buf, n)
FILE *fp;
char *buf;
int n;
{
    while(n--)
    {
	if((*buf = getc(fp)) == EOF)
	    return ERROR;
	buf++;
    }
    
    return ferror(fp);
}



/*
 * Read packet header from file
 */
int pkt_get_hdr(fp, pkt)
FILE *fp;
Packet *pkt;
{
    long val;
    struct tm t;
    int ozone, dzone;
    int cw, swap;
    char xpkt[4];
    
    pkt->from.zone = pkt->from.net = pkt->from.node = pkt->from.point = 0;
    pkt->from.domain[0] = 0;
    pkt->to  .zone = pkt->to  .net = pkt->to  .node = pkt->to  .point = 0;
    pkt->to  .domain[0] = 0;

    /* Orig node */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    pkt->from.node = val;
    /* Dest node */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    pkt->to.node = val;
    /* Year */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    t.tm_year = val - 1900;
    /* Month */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    t.tm_mon = val;
    /* Day */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    t.tm_mday = val;
    /* Hour */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    t.tm_hour = val;
    /* Minute */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    t.tm_min = val;
    /* Second */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    t.tm_sec = val;
    t.tm_wday = -1;
    t.tm_yday = -1;
    t.tm_isdst = -1;
    pkt->time = mktime(&t);
    /* Baud */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    pkt->baud = val;
    /* Version --- MUST BE PKT_VERSION (2) */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    if(val != PKT_VERSION)
	return ERROR;
    pkt->version = val;
    /* Orig net */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    pkt->from.net = val;
    /* Dest net */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    pkt->to.net = val;
    /* Prod code lo */
    if((val = getc(fp)) == ERROR)           return ERROR;
    pkt->product_l = val;
    /* Revision major */
    if((val = getc(fp)) == ERROR)           return ERROR;
    pkt->rev_maj = val;
    /* Password */
    if(pkt_get_nbytes(fp, pkt->passwd, PKT_MAXPASSWD) == ERROR)  return ERROR;
    pkt->passwd[PKT_MAXPASSWD] = 0;
    /* Orig zone (FTS-0001 optional, QMail) */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    ozone = val;
    if(ozone)
	pkt->from.zone = ozone;
    /* Dest zone (FTS-0001 optional, QMail) */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    dzone = val;
    if(dzone)
	pkt->to.zone = dzone;
    /* Spare (auxNet in FSC-0048) */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    /* Cap word byte-swapped copy */
    if((val = getc(fp)) == ERROR)           return ERROR;
    swap = val << 8;
    if((val = getc(fp)) == ERROR)           return ERROR;
    swap |= val;
    /* Prod code hi */
    if((val = getc(fp)) == ERROR)           return ERROR;
    pkt->product_h = val;
    /* Revision minor */
    if((val = getc(fp)) == ERROR)           return ERROR;
    pkt->rev_min = val;
    /* Cap word */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    cw = val;
    if(cw == swap)	/* 2+ packet according to FSC-0039 */
	debug(5, "Packet: type 2+");
    else
	cw = 0;
    pkt->capword = cw;
    /* Orig zone (FSC-0039) */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    if(cw)
	pkt->from.zone = val;
    if(ozone != val)
	debug(5, "Packet: different zones %d / %d", ozone, val);
    /* Dest zone (FSC-0039) */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    if(cw)
	pkt->to.zone = val;
    if(dzone != val)
	debug(5, "Packet: different zones %d / %d", dzone, val);
    /* Orig point (FSC-0039) */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    if(cw)
	pkt->from.point = val;
    /* Dest point (FSC-0039) */
    if((val = pkt_get_int16(fp)) == ERROR)  return ERROR;
    if(cw)
	pkt->to.point = val;
    /* Prod specific data */
    if(pkt_get_nbytes(fp, xpkt, 4) == ERROR)  return ERROR;

    debug(2, "Reading FTN packet header:");
    debug(2, "    From: %s", node_to_asc(&pkt->from, TRUE));
    debug(2, "    To  : %s", node_to_asc(&pkt->to  , TRUE));
    debug(2, "    Date: %s", date(NULL, &pkt->time));
    debug(3, "    Baud: %d", pkt->baud);
    debug(2, "    Prod: %02x %02x", pkt->product_h, pkt->product_l);
    debug(2, "    Rev : %d.%d", pkt->rev_maj, pkt->rev_min);
    debug(2, "    Pass: \"%s\"", pkt->passwd);
    debug(3, "    Capw: %04x", pkt->capword & 0xffff);
    
    return ferror(fp);
}



/*
 * Write string to packet, padded with 0 bytes to length n
 */
int pkt_put_string_padded(fp, s, n)
FILE *fp;
char *s;
int n;
{
    int i;
    for(i=0; *s && i<n; s++, i++)
	putc(*s, fp);
    for(; i<n; i++)
	putc(0, fp);
    
    return ferror(fp);
}



/*
 * Write packet header to file. This function always writes a 2+
 * (FSC-0039) header.
 */
int pkt_put_hdr(fp, pkt)
FILE *fp;
Packet *pkt;
{
    struct tm *tm;
    int swap;

    /*
     * Fill rest of Packet structure
     */
    pkt->baud      = 0;
    pkt->version   = PKT_VERSION;
    pkt->product_l = PRODUCT_CODE;
    pkt->product_h = 0;
    pkt->rev_min   = version_minor();
    pkt->rev_maj   = version_major();
    pkt->capword   = 0x0001;		/* Designates packet type 2+ */
    swap           = 0x0100;		/* Byte swapped capability word */
    tm = localtime(&pkt->time);
    

    debug(2, "Writing FTN 2+ packet header:");
    debug(2, "    From: %s", node_to_asc(&pkt->from, TRUE));
    debug(2, "    To  : %s", node_to_asc(&pkt->to  , TRUE));
    debug(2, "    Date: %s", date(NULL, &pkt->time));
    debug(3, "    Baud: %d", pkt->baud);
    debug(2, "    Prod: %02x %02x", pkt->product_h, pkt->product_l);
    debug(2, "    Rev : %d.%d", pkt->rev_maj, pkt->rev_min);
    debug(2, "    Pass: \"%s\"", pkt->passwd);
    debug(3, "    Capw: %04x", pkt->capword & 0xffff);

    /*
     * Write the actual header
     */
    pkt_put_int16        (fp, pkt->from.node);
    pkt_put_int16        (fp, pkt->to  .node);
    pkt_put_int16        (fp, tm->tm_year+1900);
    pkt_put_int16        (fp, tm->tm_mon);
    pkt_put_int16        (fp, tm->tm_mday);
    pkt_put_int16        (fp, tm->tm_hour);
    pkt_put_int16        (fp, tm->tm_min);
    pkt_put_int16        (fp, tm->tm_sec);
    pkt_put_int16        (fp, pkt->baud);
    pkt_put_int16        (fp, pkt->version);
    pkt_put_int16        (fp, pkt->from.net);
    pkt_put_int16        (fp, pkt->to  .net);
    putc                 (    pkt->product_l, fp);
    putc                 (    pkt->rev_maj,   fp);
    pkt_put_string_padded(fp, pkt->passwd, PKT_MAXPASSWD);
    pkt_put_int16        (fp, pkt->from.zone);
    pkt_put_int16        (fp, pkt->to  .zone);
    pkt_put_int16        (fp, 0 /* Spare */ );
    pkt_put_int16        (fp, swap);
    putc                 (    pkt->product_h, fp);
    putc                 (    pkt->rev_min,   fp);
    pkt_put_int16        (fp, pkt->capword);
    pkt_put_int16        (fp, pkt->from.zone);
    pkt_put_int16        (fp, pkt->to  .zone);
    pkt_put_int16        (fp, pkt->from.point);
    pkt_put_int16        (fp, pkt->to  .point);
    fputs                (    "XPKT", fp);	/* Like SQUISH */

    return ferror(fp);
}
