/*:ts=8*/
/*****************************************************************************
 * FIDOGATE --- Gateway software UNIX <-> FIDO
 *
 * $Id: rfc2fido.c,v 3.7 1994/05/02 05:36:42 mj Exp mj $
 *
 * Read mail or news from standard input and convert it to a FIDO packet.
 *
 *****************************************************************************
 * Copyright (C) 1990, 1993
 *  _____ _____
 * |     |___  |   Martin Junius             FIDO:      2:242/6.1
 * | | | |   | |   Republikplatz 3           Internet:  mj@sungate.fido.de
 * |_|_|_|@home|   D-52072 Aachen, Germany   Phone:     ++49-241-86931 (voice)
 *
 * This file is part of FIDOGATE.
 *
 * FIDOGATE is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * FIDOGATE is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with FIDOGATE; see the file COPYING.  If not, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************/

#include "fidogate.h"
#include "getopt.h"

#include <unistd.h>
#include <pwd.h>
#include <sys/stat.h>



#define PROGRAM "rfc2fido"
#define VERSION "$Revision: 3.7 $"


/*
 * Prototypes
 */
char *new_packet_name	P((char *));
int   new_packet	P((char *));
void  close_packet	P((int));
void  sendback		P((const char *, ...));
char *mail_sender	P((Node *));
char *mail_address	P((void));
char *receiver		P((char *, Node *));
char *mail_receiver	P((char *, Node *));
time_t	mail_date		P((void));
char *estrtok		P((char *, char *));
int snd_mail		P((char *, char *, int));
int snd_message		P((Message *,Area *,char *,char *,int,char *,int));
int print_tear_line	P((FILE *));
int print_origin	P((FILE *, char *));
int print_local_msgid	P((FILE *));
int print_via		P((FILE *));

void short_usage	P((void));
void usage		P((void));



/*
 * Max. message size for FIDO. Due to some more brain damage in FIDONET
 * programs we have to split larger messages into several smaller ones.
 * We use 13000 bytes as the max. message size, this leaves enough room for
 * headers and other things.
 */
#define MAXMSGSIZE	13000



/*
 * Outbound packets: directory, name, temp. name, file
 */
char  packet_dir [MAXPATH];
char  packet_name[MAXPATH];
char  packet_tmp [MAXPATH];
FILE *packet_file = NULL;
#ifdef BINKLEY_FLO
char  flo_name   [MAXPATH];
FILE *flo_file;
#endif

/*
 * File name from -o command line option
 */
char out_name[MAXPATH] = "";

/*
 * Zone address of packet written last to (== 0 no packet yet)
 */
int last_zone = 0;


/* Private mail (default) */
bool private = TRUE;

/* News-article */
int newsmode = FALSE;




/*
 * Global Textlist to save message body
 */
Textlist body = { NULL, NULL };



/*
 * Get name for new packet
 */
char *new_packet_name(name)
char *name;
{
    if(name)
    {
	strncpy0(packet_name, name, sizeof(packet_name));
	strncpy0(packet_tmp , name, sizeof(packet_tmp ));
    }
    else
    {
	long n;
	
	n = sequencer(SEQ_PKT);
	sprintf(packet_name, "%s/%08ld.pkt", packet_dir, n);
	sprintf(packet_tmp , "%s/%08ld.tmp", packet_dir, n);
    }

#ifdef BINKLEY_FLO
    sprintf(flo_name, "%s/%04x%04x%s", packet_dir, cf_uplink()->net,
	    cf_uplink()->node, BINKLEY_FLO);
#endif

    return packet_name;
}



/*
 * Create new packet file
 */
int new_packet(name)
char *name;
{
    Packet pkt;

    if(!name[0])
	name = NULL;
    
    if(name==NULL || packet_file==NULL)
    {
	/*
	 * Open new packet file. If a packet name is specified via -o
	 * then only create a new file on the first call to new_packet().
	 */
	new_packet_name(name);
	if((packet_file = fopen(packet_tmp, W_MODE)) == NULL)
	{
	    log("$%s: can't create packet %s: ", PROGRAM, packet_tmp);
	    exit(EX_CANTCREAT);
	}

	/*
	 * Change mode to PACKET_MODE
	 */
	chmod(packet_tmp, PACKET_MODE);

	debug(1, "New packet file %s (tmp %s)", packet_name, packet_tmp);
	
	/*
	 * Write packet header
	 */
	pkt.from = cf_n_fake();
	pkt.to   = cf_n_uplink();
	pkt.time = time(NULL);
	strncpy0(pkt.passwd, PACKET_PASSWD, sizeof(pkt.passwd));
	/* Rest is filled in by pkt_put_hdr() */
	if(pkt_put_hdr(packet_file, &pkt) == ERROR)
	{
	    log("$Can't write to packet file %s", packet_tmp);
	    exit(EX_OSERR);
	}
    }

    /*
     * packet_file==NULL can't happen here, but anyway ...
     */
    return packet_file ? ferror(packet_file) : ERROR;
}



/*
 * Write end of packet to packet file, close and rename it.
 */
void close_packet(force)
int force;
{
    if(force || !out_name[0])
	/*
	 * Only close packet, if force==TRUE or no -o option given
	 */
	if(packet_file)
	{
	    /* End of packet */
	    pkt_put_int16(packet_file, 0);
	    fclose(packet_file);
	    
	    /* Rename .tmp -> .pkt */
	    if(!out_name[0])
		if(rename(packet_tmp, packet_name) == ERROR)
		    log("$Can't rename %s to %s", packet_tmp, packet_name);
#ifdef BINKLEY_FLO
		else
		{
		    flo_file = fopen(flo_name, A_MODE);
		    chmod(flo_name, PACKET_MODE);
		    fprintf(flo_file, "^%s\n", packet_name);
		    fclose(flo_file);
		}
#endif
	}
}



/*
 * In case of error print message (mail returned by MTA)
 */

#ifdef __STDC__
void sendback(const char *fmt, ...)
{
    va_list args;
    
    va_start(args, fmt);
#else /**!__STDC__**/    
void sendback(va_alist)
va_dcl
{
    va_list args;
    char *fmt;

    va_start(args);
    fmt = va_arg(args, char *);
#endif /**__STDC__**/

    fprintf(stderr, "Internet -> FIDO gateway / FIDOGATE %s\n",
	    version_global()                                   );
    fprintf(stderr, "   ----- ERROR -----\n");
    vfprintf(stderr, fmt, args);
    fprintf(stderr, "\n");
}



/*
 * Return sender's full name. This is taken from
 *   - Reply-To header
 *   - From header
 *   - user id and password file
 *   - environment LOGNAME or USER
 */

char *mail_sender(node)
    Node *node;
{
    struct passwd *pwd;
    static char buffer[BUFSIZ];
#ifdef PASSTHRU_NETMAIL
    static char name[MSG_MAXNAME];
    Node n;
#endif /**PASSTHRU_NETMAIL**/
    char *from;
    char *p;
    
    /*
     * Default sender node is gateway address
     */
    *node = cf_n_addr();
    
    /*
     * Look up name in Reply-To or From header of message
     */
    if((from = header_getcomplete("Reply-To")) == NULL)
	from = header_getcomplete("From");

    if(from)
    {
	/*
	 * Found Reply-To or From header
	 */
	debug(3, "RFC From: %s", from);

#ifdef PASSTHRU_NETMAIL
	/*
	 * If the from address is an FTN address, convert and pass it via
	 * parameter node. This may cause problems when operating different
	 * FTNs.
	 */
	if(addr_from_rfcaddr(from, buffer))
	    if(parse_address(buffer, name, &n) != ERROR)
	    {
		*node = n;
		debug(3, "    is FTN address %s", node_to_asc(node, TRUE));
	    }
#endif /**PASSTHRU_NETMAIL**/

	if(!name_from_rfcaddr(from, buffer))
	    username_from_rfcaddr(from, buffer);
	debug(3, "RFC Full name: %s", buffer);
	return buffer;
    }

    /*
     * Try user id and passwort file entry
     */
    if((pwd = getpwuid(getuid())))
    {
	strcpy(buffer, pwd->pw_gecos);
	if((p = strchr(buffer, ',')))
	    /*
	     * Kill stuff after ','
	     */
	    *p = 0;
	if(!*buffer)
	    /*
	     * Empty, use user name
	     */
	    strcpy(buffer, pwd->pw_name);
	debug(3, "passwd Full name: %s", buffer);
	return buffer;
    }

    /*
     * Use user name from environment as a last resort
     */
    if((p = getenv("LOGNAME")))
    {
	strcpy(buffer, p);
	debug(3, "Env LOGNAME Full name: %s", buffer);
	return buffer;
    }
    if((p = getenv("USER")))
    {
	strcpy(buffer, p);
	debug(3, "Env USER Full name: %s", buffer);
	return buffer;
    }

    debug(5, "No name for sender found");
    strcpy(buffer, "Unknown User");
    return buffer;
}



/*
 * Return sender's mail address. This is taken from
 *   - Reply-To header
 *   - From header
 *   - user id and password file
 *   - environment LOGNAME or USER
 */

char *mail_address()
{
    struct passwd *pwd;
    char *p;
    
    /*
     * Look up name in Reply-To or From header of message
     */
    if((p = header_getcomplete("Reply-To")) == NULL)
	p = header_getcomplete("From");

    if(p)
	return buf_copy(p);

    /*
     * Try user id and passwort file entry
     */
    if((pwd = getpwuid(getuid())))
	return buf_sprintf("%s@%s", pwd->pw_name, cf_fqdn() );

    /*
     * Use user name from environment as a last resort
     */
    if((p = getenv("LOGNAME")))
	return buf_sprintf("%s@%s", p, cf_fqdn() );
    if((p = getenv("USER")))
	return buf_sprintf("%s@%s", p, cf_fqdn() );

    return NULL;
}



/*
 * receiver() --- Check for aliases and beautify name
 */

char *receiver(to, node)
    char *to;
    Node *node;
{
    static char name[MSG_MAXNAME];
    int i, c, convert_flag;
    Alias *alias;

    /*
     * Check for name alias
     */
    debug(5, "Name for alias checking: %s", to);

    if((alias = alias_lookup(node, to, NULL)))
    {
	debug(5, "Alias found: %s %s %s", alias->username,
	      node_to_asc(&alias->node, TRUE), alias->fullname);
	strcpy(name, alias->fullname);

	/*
	 * Store address from ALIASES into node, this will reroute
	 * the message to the point specified in ALIASES, if the message
	 * addressed to node without point address.
	 */
	*node = alias->node;
	
	return name;
    }
    
    /*
     * Alias not found. Return the the original receiver with all
     * '_' characters replaced by space and all words capitalized.
     */
    convert_flag = isupper(*to) ? -1 : 1;
    for(i=0; *to && i<35; i++, to++) {
	c = *to;
	switch(c) {
	    case '_':
		name[i] = ' ';
		if(!convert_flag)
		    convert_flag = 1;
		break;
	    case '%':
		if(convert_flag != -1)
		    convert_flag = 2;
		/**Fall thru**/
	    default:
		if(convert_flag > 0) {
		    name[i] = islower(c) ? toupper(c) : c;
		    if(convert_flag == 1)
			convert_flag = 0;
		}
		else
		    name[i] = c;
		break;
	}
    }
    name[i] = 0;

    debug(5, "No alias found: return %s", name);

    return name;
}



/*
 * Return from field for FIDO message.
 * Alias checking is done via receiver().
 */

char *mail_receiver(address, node)
char *address;
Node *node;
{
    char *to;
    char name[MSG_MAXNAME];
    char realname[128];

#if 0 /**See below**/
    char *p;
    int found = 0;
    char addr[128], myaddr[128];
    Node dummy;
#endif /**0**/

    realname[0] = 0;

    if(address) {
	/*
	 * Address is argument
	 */
	if(parse_address(address, name, node)) {
	    log("Parse of address %s failed", address);
	    return NULL;
	}
    }
    else {
	/*
	 * News/EchoMail: address is echo feed
	 */
	*node = cf_n_uplink();
	strcpy(name, "All");
    
	/*
	 * User-defined header line X-Comment-To for gateway software
	 * (can be patched into news reader)
	 */
	if( (to = header_get("X-Comment-To")) )
	    if(name_from_rfcaddr(to, realname) && *name)
		strncpy0(name, realname, sizeof(name));
    }

#if 0 /**Disabled checking of headers for real names**/
    /*
     * Try to look up a better real name in header fields
     *
     * Standard RFC822 header line
     */
    if(to = header_get("To")) {
	debug(5, "Checking To: field");
	addr_from_rfcaddr(to, addr);
	if(!address || !strcmp(address, addr)) {
	    name_from_rfcaddr(to, realname);
	    found = 1;
	}
    }

    /*
     * User-defined header line for gateway software
     * (can be patched into news reader)
     */
    if(!found && (to = header_get("X-Comment-To"))) {
	debug(5, "Checking X-Comment-To: field");
	addr_from_rfcaddr(to, addr);
	if(!address || !strcmp(address, addr)) {
	    name_from_rfcaddr(to, realname);
	    found = 1;
	}
    }
    
    /*
     * Header generated by nn's `r' command
     */
    if(!found && (to = header_get("Orig-To"))) {
	debug(5, "Checking Orig-To: field");
	addr_from_rfcaddr(to, addr);
	if(!address || !strcmp(address, addr)) {
	    name_from_rfcaddr(to, realname);
	    found = 1;
	}
    }
    
    /*
     * News message: get name from address taken out of header line
     */
    if(found && !address) {
	/*
	 * If we've got `user%p.f.n.z@HOST.DOMAIN' address, strip
	 * of `@HOST.DOMAIN' part.
	 */
	if((p = strrchr(addr, '@'))) {
	    strcpy(myaddr, cf_hostname() );
	    if(!stricmp(p+1, myaddr))
		*p = 0;
	    strcat(myaddr, cf_domainname() );
	    if(!stricmp(p+1, myaddr))
		*p = 0;
	}
	    
	parse_address(addr, name, &dummy);
    }
	
    /*
     * Use real name from header line, if no special addressing with '%'
     */
    if(found && !strchr(name, '%') && *realname)
	strncpy0(name, realname, MSG_MAXNAME);
#endif /**0**/
    
    return receiver(name, node);
}



/*
 * Get date field for FIDO message. Look for `Date:' header or use
 * current time.
 */
time_t mail_date()
{
    time_t timevar = -1;
    char *header_date;

    if((header_date = header_get("Date"))) {
	/* try to extract date and other information from it */
	debug(3, "RFC Date: %s", header_date);
	timevar = parsedate(header_date, NULL);
	if(timevar == -1)
	    log("Can't parse Date: %s", header_date);
    }

    return timevar;
}



/*
 * Process mail/news message
 */
int snd_mail(from, to, split)
    char *from;				/* Address from */
    char *to;				/* Address to send to (news = NULL) */
    int split;				/* Split message into several parts */
{
    char groups[BUFSIZ];
    Node node_from, node_to;
    char *p;
    int status, fido;
    Message msg;
    char *flags;
    
    
    node_from.domain[0] = 0;
    node_to  .domain[0] = 0;
    
    if(to)
	debug(3, "RFC To: %s", to);

    /*
     * From name/node
     */
    p = mail_receiver(to, &node_to);
    if(!p) {
	if(*address_error)
	    sendback("Address %s:\n  %s", to, address_error);
	else
	    sendback("Address %s:\n  address/host is unknown", to);
	return(EX_NOHOST);
    }
    strncpy0(msg.name_to, p, sizeof(msg.name_to));
    fido = isfido();

    cf_set_zone(node_to.zone);

    /*
     * To name/node
     */
    p = mail_sender(&node_from);
    if(!p)
	p = "Gateway";
    strncpy0(msg.name_from, p, sizeof(msg.name_from));

    /*
     * Subject
     */
    p = header_get("Subject");
    if(!p)
	p = "(no subject)";
    if(split)
    {
	strcpy(msg.subject, "X: ");
	strncpy0(msg.subject+3, p, sizeof(msg.subject)-3);
    }
    else
	strncpy0(msg.subject, p, sizeof(msg.subject));

    /*
     * Date
     */
    msg.date = mail_date();
    
    msg.cost = 0;
    msg.attr = 0;
    if(private)
	msg.attr |= MSG_PRIVATE;
    flags = p = header_get("X-Flags");
    if(p)
    {
	while(*p)
	    switch(*p++)
	    {
	    case 'c': case 'C':
		msg.attr |= MSG_CRASH;
		break;
	    case 'p': case 'P':
		msg.attr |= MSG_PRIVATE;
		break;
	    case 'h': case 'H':
		msg.attr |= MSG_HOLD;
		break;
	    }
    }	
		
    if(newsmode) {
	Area *pa;
	int xpost_flag;
	
	/*
	 * News message: get newsgroups and convert to FIDO areas
	 */
	p = header_get("Newsgroups");
	if(!p) {
	    sendback("No Newsgroups header in news message");
	    return(EX_DATAERR);
	}
	strncpy0(groups, p, sizeof(groups));
	debug(3, "RFC Newsgroups: %s", groups);

	xpost_flag = strchr(groups, ',') != NULL;
	
	for(p=strtok(groups, ","); p; p=strtok(NULL, ","))
	{
	    debug(5, "Look up newsgroup %s", p);
	    pa = areas_lookup(NULL, p);
	    if(!pa)
		debug(5, "No FTN area");
	    else
	    {
		debug(5, "Found: %s %s Z%d", pa->area, pa->group, pa->zone);

		if( xpost_flag && (pa->flags & AREA_NOXPOST) )
		{
		    debug(5, "No Xpostings allowed - skipped");
		    continue;
		}
		if( xpost_flag && (pa->flags & AREA_LOCALXPOST) )
		{
		    if(addr_is_local(from))
		    {
			debug(5, "Local Xposting - o.k.");
		    }
		    else
		    {
			debug(5, "No non-local Xpostings allowed - skipped");
			continue;
		    }
		}

		cf_set_zone(pa->zone);
		msg.area      = pa->area;
		msg.node_from = cf_n_fake();
		msg.node_to   = cf_n_uplink();
		status = snd_message(&msg, pa, from,
				     NULL, split, flags, fido);
		if(status)
		    return status;
	    }
	}
    }
    else {
	/*
	 * NetMail message
	 */
	log("%-13s  %s -> %s", node_to_asc(&node_to, TRUE), from, to);
	msg.area      = NULL;
	msg.node_from = node_from;
	msg.node_to   = node_to;
	return snd_message(&msg, NULL, from, to, split, flags, fido);
    }
    
    return EX_OK;
}


int snd_message(msg, parea, from, to, split, flags, fido)
    Message *msg;			/* FTN nessage structure */
    Area *parea;			/* Area/newsgroup */
    char *from;				/* Internet sender */
    char *to;				/* Internet recipient */
    int split;				/* >0: splitted message part */
    char *flags;			/* From X-Flags header */
    int fido;				/* TRUE: recipient is FTN address */
{
    FILE *sf;
    char *header;
    int part = 1, line = 1;
    long size;
    Textline *p;
    char *id;
    int flag;
    
    /*
     * Set pointer to first line in message body
     */
    p = body.first;
    
    /*
     * Output packet
     */
    if(cf_zone() != last_zone)
    {
	/*
	 * Create new packet, if this message is generated with a different
	 * zone AKA.
	 */
	close_packet(FALSE);
	new_packet(out_name);
	last_zone = cf_zone();
    }
    
    sf = packet_file;

 again:

    /* Subject with splitted part indication */
    if(split)
	msg->subject[0] = part + '@';

    /* Header */
    pkt_put_msg_hdr(sf, msg, TRUE);


    if(!flags || !strchr(flags, 'M'))		/* X-Flags: M */
    {
	/*
	 * Add kludges for MSGID / REPLY and ORIGID / ORIGREF
	 */
	if((header = header_get("Message-ID")))
	{
	    if((id = msgid_rfc_to_fido(&flag, header, part, split)))
	    {
		fprintf(sf, "\001MSGID: %s\r\n", id);
		if(flag && (id = msgid_rfc_to_origid(header, part, split)))
		    fprintf(sf, "\001ORIGID: %s\r\n", id);
	    }
	    else
		print_local_msgid(sf);
	}	
	else
	    print_local_msgid(sf);
	
	if((header = header_get("References")) ||
	   (header = header_get("In-Reply-To")))
	{
	    if((id = msgid_rfc_to_fido(&flag, header, 0, 0)))
	    {
		fprintf(sf, "\001REPLY: %s\r\n", id);
		if(flag && (id = msgid_rfc_to_origid(header, 0, 0)))
		    fprintf(sf, "\001ORIGREF: %s\r\n", id);
	    }
	}
    }
    else
	print_local_msgid(sf);

#ifdef FSC_0035
    if(!flags || !strchr(flags, 'N'))
    {
	/*
	 * Generate FSC-0035 ^AREPLYADDR, ^AREPLYTO
	 */
	if(from) 
	{
	    fprintf(sf, "\001REPLYADDR %s\r\n", from);
	    fprintf(sf, "\001REPLYTO %s %s\r\n",
		    node_to_asc(cf_addr(), FALSE),
		    msg->name_from);
	}
    }
#endif /**FSC_0035**/    

    /*
     * Add ^ACHRS kludge indicating that this messages uses the ISO-8859-1
     * charset. This may be not the case, but most Usenet messages with umlauts
     * use this character set and FIDOGATE doesn't support MIME yet.
     * (^ACHRS is documented in FSC-0054)
     */
    fprintf(sf, "\001CHRS: LATIN-1 2\r\n");


#ifdef ROUTE_VIA_FTN_GATEWAY
    /*
     * Add To line for addressing FIDO<->Internet gateway
     */
    if(to && !fido)
	fprintf(sf, "To: %s\r\n", to);
#endif /**ROUTE_VIA_FTN_GATEWAY**/

    if(!flags || !strchr(flags, 'N'))
    {
#ifndef NO_FROM_LINE
	/*
	 * Add From line with return address
	 */
	if(from)
	    fprintf(sf, "From: %s\r\n", from);
#endif /**NO_FROM_LINE**/
	/*
	 * Add some header lines as info for user
	 */
	if((header = header_get("From")))
	{
	    if(strcmp(from, header)) 
	    {
		/*
		 * From header line is not equal to reply address, output
		 * From and Reply-To header line.
		 */
		fprintf(sf, " * From: %s\r\n", header);
		while((header = header_getnext()))
		    fprintf(sf, " *       %s\r\n", header);
		if((header = header_get("Reply-To")))
		{
		    fprintf(sf, " * Reply-To: %s\r\n", header);
		    while((header = header_getnext()))
			fprintf(sf, " *           %s\r\n", header);
		}
	    }		    
	}
	if((header = header_get("To")))
	{
	    fprintf(sf, " * To: %s\r\n", header);
	    while((header = header_getnext()))
		fprintf(sf, " *     %s\r\n", header);
	}
	if((header = header_get("Cc")))
	{
	    fprintf(sf, " * Cc: %s\r\n", header);
	    while((header = header_getnext()))
		fprintf(sf, " *     %s\r\n", header);
	}
	if((header = header_get("Newsgroups")))
	    if(strchr(header, ',') || !newsmode)
		/* Posted to multiple groups or mail */
		fprintf(sf, " * Newsgroups: %s\r\n", header);
	fprintf(sf, "\r\n");
    }
	
    /*
     * Add line indicating splitted message
     */
    if(split)
	fprintf(sf,
		" * Large message splitted by FIDOGATE: part %02d/%02d\r\n\r\n",
		part, split						    );

    /*
     * Copy message body
     */
    size = 0;
    debug(5, "Copying message body");
    while(p)
    {
#if 0 /**DISABLED**/
	/*
	 * BEWARE! This is a kludge!
	 * The stripping of signatures should be configurable in the
	 * `areas' file. Currently it is applied for all fido.ger.* groups.
	 * Also the max length of a signature is somewhat arbitrary.
	 *
	 * ----- CHANGE ME !!! -----
	 */
	if(pa && pa->group && !strncmp(pa->group, "fido.ger.", 9)
	   /* Only for fido.ger.* groups */
	   && ( !strcmp(p->line, "--\n") || !strcmp(p->line, "-- \n") )
	   /* Signature marks */
	   && (body.n - line) < 15
	   /* Assumed max length of a signature, avoid deleting real text */)
	{
	    debug(4, "Deleting signature");
/*	    put_line(sf, "-- SIGNATURE DELETED --\n"); */
	    break;
	}
#endif
	
	pkt_put_line(sf, p->line);
	size += strlen(p->line) + 1;
	if(size > MAXMSGSIZE) {
	    print_tear_line(sf);
	    if(newsmode)
		print_origin(sf,
			     parea && parea->origin ? parea->origin
			     : cf_origin() );
	    /* End of message */
	    putc(0, sf);
	    part++;
	    p = p->next;
	    goto again;
	}
	p = p->next;
	line++;
    }

    /*
     * If message is for echo mail (-n flag) then add
     * tear, origin, seen-by and path line.
     */
    print_tear_line(sf);
    if(newsmode)
	print_origin(sf,
		     parea && parea->origin ? parea->origin : cf_origin() );
    else
	print_via(sf);

    /*
     * Done
     */
    /* End of message */
    putc(0, sf);
    return EX_OK;
}



/*
 * Output tear line
 */
int print_tear_line(fp)
FILE *fp;
{
    fprintf(fp, "\r\n--- FIDOGATE %s\r\n", version_global());

    return ferror(fp);
}



/*
 * Generate origin, seen-by and path line
 */
int print_origin(fp, origin)
    FILE *fp;
    char *origin;
{
    char buf[80];
    char bufa[30];
    int len;
    
    strncpy0(buf, " * Origin: ", sizeof(buf));
    strncpy0(bufa, node_to_asc(cf_addr(), TRUE), sizeof(bufa));
    
    /*
     * Max. allowed length of origin line is 79 (80 - 1) chars,
     * 3 chars for additional " ()".
     */
    len = 80 - strlen(bufa) - 3;

    /* Add origin text */
    strncat0(buf, origin, len);
    /* Add address */
    strncat0(buf, " (", sizeof(buf));
    strncat0(buf, bufa, sizeof(buf));
    strncat0(buf, ")" , sizeof(buf));
    
    /* Origin */
    fprintf(fp, "%s\r\n", buf);

    if(cf_fake()->point)		/* Generate 4D addresses */
    {
	/* SEEN-BY */
	fprintf(fp, "SEEN-BY: %d/%d",
		cf_fake()->net, cf_fake()->node);
#ifdef FTN_4D_ECHOMAIL_WITH_POINTS
	fprintf(fp, ".%d", cf_fake()->point);
#endif
	/* PATH */
	fprintf(fp, "\r\n\001PATH: %d/%d",
		cf_fake()->net, cf_fake()->node);
#ifdef FTN_4D_ECHOMAIL_WITH_POINTS
	fprintf(fp, ".%d", cf_fake()->point);
#endif
	fputs("\r\n", fp);
    }
    else				/* Generate 3D addresses */
    {
	/* SEEN-BY */
	fprintf(fp, "SEEN-BY: %d/%d ",
		cf_fake()->net, cf_fake()->node );
	if(cf_uplink()->net != cf_fake()->net)
	    fprintf(fp, "%d/", cf_uplink()->net);
	fprintf(fp,"%d\r\n", cf_uplink()->node);
	/* PATH */
	fprintf(fp, "\001PATH: %d/%d\r\n",
		cf_fake()->net, cf_fake()->node);
    }
    
    return ferror(fp);
}



/*
 * Generate local `^AMSGID:' if none is found in message header
 */
int print_local_msgid(fp)
FILE *fp;
{
    long msgid;

    msgid = sequencer(SEQ_MSGID);
    fprintf(fp, "\001MSGID: %s %08ld\r\n",
	    node_to_asc(cf_addr(), FALSE), msgid);

    return ferror(fp);
}



/*
 * Generate "^AVia" line for NetMail
 */
int print_via(fp)
FILE *fp;
{
    fprintf(fp, "\001Via FIDOGATE %s, %s\r\n",
	    node_to_asc(cf_addr(), TRUE),
	    date("%a %b %d %Y at %H:%M:%S %Z", NULL)  );

    return ferror(fp);
}



/*
 * Usage messages
 */
void short_usage()
{
    fprintf(stderr, "usage: %s [-options] [user ...]\n", PROGRAM);
    fprintf(stderr, "       %s --help  for more information\n", PROGRAM);
}


void usage()
{
    fprintf(stderr, "FIDOGATE %s  %s %s\n\n",
	    version_global(), PROGRAM, version_local(VERSION) );
    
    fprintf(stderr, "usage:   %s [-options] [user ...]\n\n", PROGRAM);
    fprintf(stderr, "\
options: -b --news-batch              process news batch\n\
	 -o --out-packet-file  NAME   set outbound packet file name\n\
	 -n --news-mode               set news mode\n\
         -i --ignore-registration     do not bounce unknown p.f.n.z.fido.de\n\
	 -O --out-packet-dir   NAME   set outbound packet directory\n\
\n\
	 -v --verbose                 more verbose\n\
	 -h --help                    this help\n\
         -c --config name             read config file (\"\" = none)\n\
	 -L --lib-dir name            set lib directory\n\
	 -S --spool-dir name          set spool directory\n\
	 -a --addr Z:N/F.P            set FTN address\n\
	 -f --fake-addr Z:N/F.P       set FTN fakenet address\n\
	 -u --uplink-addr Z:N/F.P     set FTN uplink address\n\
         -z --zone Z                  select aka for zone Z\n"  );
}



/***** main ******************************************************************/
int main(argc, argv)
    int argc;
    char *argv[];
{
    int cnt, c;
    int status = EX_OK;
    long size, nmsg;
    int split;
    char *from;
    int b_flag=FALSE;
    char *O_flag=NULL;
    char *c_flag=NULL;
    char *S_flag=NULL, *L_flag=NULL;
    char *a_flag=NULL, *f_flag=NULL, *u_flag=NULL;
    char *z_flag=NULL;
    int option_index;
    static struct option long_options[] =
    {
	{ "news-batch",   0, 0, 'b'},	/* Process news batch */
	{ "out-packet-file",1,0,'o'},	/* Set packet file name */
	{ "news-mode",    0, 0, 'n'},	/* Set news mode */
	{ "ignore-registration", 0, 0, 'i'}, /* Do not bounce unknown fido.de*/
	{ "out-dir",      1, 0, 'O'},	/* Set packet directory */

	{ "verbose",      0, 0, 'v'},	/* More verbose */
	{ "help",         0, 0, 'h'},	/* Help */
	{ "config",       1, 0, 'c'},	/* Config file */
	{ "spool-dir",    1, 0, 'S'},	/* Set FIDOGATE spool directory */
	{ "lib-dir",      1, 0, 'L'},	/* Set FIDOGATE lib directory */
	{ "addr",         1, 0, 'a'},	/* Set FIDO address */
	{ "fake-addr",    1, 0, 'f'},	/* Set FIDO fakenet address */
	{ "uplink-addr",  1, 0, 'u'},	/* Set FIDO uplink address */
	{ "zone",         1, 0, 'z'},	/* Select aka for zone */
	{ 0,              0, 0, 0  }
    };

    /* Init configuration */
    cf_initialize();
    
    newsmode = FALSE;

    

    while ((c = getopt_long(argc, argv, "bo:niO:vhc:L:S:a:f:u:z:",
			    long_options, &option_index     )) != EOF)
	switch (c) {
	/***** rfc2fido options *****/
	case 'b':
	    /* News batch flag */
	    b_flag   = TRUE;
	    newsmode = TRUE;
	    private  = FALSE;
	    break;
	case 'o':
	    /* Set packet file name */
	    strncpy0(out_name, optarg, sizeof(out_name));
	    break;
	case 'n':
	    /* Set news-mode */
	    newsmode = TRUE;
	    private  = FALSE;
	    break;
	case 'i':
	    /* Don't bounce non-Fido.DE */
	    i_flag = TRUE;
	    break;
	case 'O':
	    /* Set packet dir */
	    O_flag = optarg;
	    break;

	/***** Common options *****/
	case 'v':
	    verbose++;
	    break;
	case 'h':
	    usage();
	    exit(0);
	    break;
	case 'c':
	    c_flag = optarg;
	    break;
	case 'S':
	    S_flag = optarg;
	    break;
	case 'L':
	    L_flag = optarg;
	    break;
	case 'a':
	    a_flag = optarg;
	    break;
	case 'f':
	    f_flag = optarg;
	    break;
	case 'u':
	    u_flag = optarg;
	    break;
	case 'z':
	    z_flag = optarg;
	    break;
	default:
	    short_usage();
	    exit(EX_USAGE);
	    break;
	}


    /*
     * Read config file
     */
    if(L_flag)				/* Must set libdir beforehand */
	cf_set_libdir(L_flag);
    cf_read_config_file(c_flag);

    /*
     * Process config options
     */
    if(L_flag)
	cf_set_libdir(L_flag);
    if(S_flag)
	cf_set_spooldir(S_flag);
    if(a_flag)
	cf_set_addr(a_flag);
    if(f_flag)
	cf_set_fake(f_flag);
    if(u_flag)
	cf_set_uplink(u_flag);

    cf_debug();

    if(z_flag)
	cf_set_zone(atoi(z_flag));
    
    /*
     * Process local options
     */
    if(O_flag)
	strncpy0(packet_dir, O_flag, sizeof(packet_dir));
    else
	sprintf(packet_dir, "%s/%s", cf_spooldir(), OUTBOUND);

    /*
     * Init various modules
     */
    if(newsmode)
	areas_init();
    hosts_init();
    alias_init();
#ifdef MAUS_DE
    maus_init();
#endif

    /**
     ** Main loop: read message(s) from stdin, batches if -b
     **/
    nmsg = 0;
    while(TRUE)
    {
	if(b_flag)
	{
	    size = read_rnews_size(stdin);
	    if(size == -1)
		log("%s: error reading news batch", PROGRAM);
	    if(size <= 0)
		break;
	    nmsg++;

	    debug(1, "Batch: message #%ld size %ld", nmsg, size);
	}
	
	/*
	 * Read message header from stdin
	 */
	header_delete();
	header_read(stdin);
    
	/*
	 * Read message body from stdin and count size
	 */
	size = 0;
	Textlist_clear(&body);
	while(read_line(buffer, BUFFERSIZE, stdin)) {
	    Textlist_append(&body, buffer);
	    size += strlen(buffer) + 1;	    /* `+1' for additional CR */
	}
	split = size>MAXMSGSIZE ? size/MAXMSGSIZE + 1 : 0;
	debug(1, "Message body size %ld", size);
	if(split)
	    debug(1, "Must split message, %d parts", split);
	from = mail_address();		/* From address */

	if(newsmode)
	    /*
	     * Send mail to echo feed for news messages
	     */
	    status = snd_mail(from, NULL, split);
	else
	    /*
	     * Send mail to addresses from command line args
	     */
	    for(cnt = optind; cnt < argc; cnt++)
		if((status = snd_mail(from, argv[cnt], split)) != EX_OK)
		    break;

	if(!b_flag)
	    break;
    }

    close_packet(TRUE);
    
    exit(status);
}
